# -*- coding: utf-8

"""
This module is designed to hold optimization model implementations. The
implementation makes use of oemof-solph.

This file is part of project dhnx (). It's copyrighted
by the contributors recorded in the version control history of the file,
available from its original location:

SPDX-License-Identifier: MIT
"""

import os
import logging
import networkx as nx
import pandas as pd
import oemof.solph as solph
from oemof.solph import helpers
from oemof.tools import logger

from .optimization_dhs_nodes import add_nodes_dhs, add_nodes_houses
from .model import OperationOptimizationModel, InvestOptimizationModel


class OemofOperationOptimizationModel(OperationOptimizationModel):
    r"""
    Implementation of an operation optimization model using oemof-solph.
    """
    def __init__(self, thermal_network):
        super().__init__(thermal_network)
        self.results = {}

    def setup(self):
        pass

    def solve(self):
        pass

    def get_results(self):
        return self.results


class OemofInvestOptimizationModel(InvestOptimizationModel):
    """
    Implementation of an invest optimization model using oemof-solph.

    ...

    Attributes
    ----------
    settings : dict
        Dictionary holding the optimisation settings. See .
    invest_options : dict
        Dictionary holding the investment options for the district heating system.
    nodes : list
        Empty list for collecting all oemof.solph nodes.
    buses : dict
        Empty dictionary for collecting all oemof.solph.Buses of the energy system.
    es : oemof.solph.EnergySystem
        Empty oemof.solph.EnergySystem.
    om : oemof.solph.Model
        Attribute, which will be the oemof.solph.Model for optimisation.
    oemof_flow_attr : set
        Possible flow attributes, which can be used additionally:
        {'nominal_value', 'min', 'max', 'variable_costs', 'fix'}
    results : dict
        Empty dictionary for the results.

    Methods
    -------
    check_input():
        Performs checks on the input data.
    complete_exist_data():
        Sets the investment status for the results dataframe of the pipes.
    get_pipe_data():
        Adds heat loss and investment costs to pipes dataframe.
    setup_oemof_es():
        The energy system *es* is build.
    setup():
        Calls *check_input()*, *complete_exist_data()*, *get_pipe_data()*, and *setup_oemof_es()*.

    """
    def __init__(self, thermal_network, settings, investment_options):

        self.settings = settings
        self.invest_options = investment_options
        self.nodes = []  # list of all nodes
        self.buses = {}  # dict of all buses
        self.es = solph.EnergySystem()
        self.om = None

        # list of possible oemof flow attributes, e.g. for producers source
        self.oemof_flow_attr = {'nominal_value', 'min', 'max',
                                'variable_costs', 'fix'}

        super().__init__(thermal_network)
        self.results = {}

    def check_input(self):
        """Check 1:

        Check and make sure, that the dtypes of the columns of the sequences
        and the indices (=ids) of the forks, pipes, producers and consumers
        are of type 'str'. (They need to be the same dtye.)

        Check 2:

        Firstly, it is checked, if there are any not-allowed connection in the *pipe* data.
        The following connections are not allowed:

          * consumer -> consumer
          * producer -> producer
          * producer -> consumer
          * consumer -> fork

        Secondly, it is checked, if a pipes goes to a consumer, which does not exist.

        Check 3

        Checks if graph of network is connected.

        An error is raised if one of these connection occurs.
        """

        # Check 1
        # make sure that all ids are of type str
        # sequences
        sequ_items = self.thermal_network.sequences.keys()
        for it in sequ_items:
            for v in self.thermal_network.sequences[it].values():
                v.columns.astype('str')

        # components
        for comp in ['pipes', 'consumers', 'producers', 'forks']:
            self.thermal_network.components[comp].index = \
                self.thermal_network.components[comp].index.astype('str')

        # Check 2

        ids_consumers = self.thermal_network.components['consumers'].index

        for p, q in self.thermal_network.components['pipes'].iterrows():

            if (q['from_node'].split('-')[0] == "consumers") and (
                    q['to_node'].split('-')[0] == "consumers"):

                raise ValueError(
                    ""
                    "Pipe id {} goes from consumer to consumer. This is not "
                    "allowed!".format(p))

            if (q['from_node'].split('-')[0] == "producers") and (
                    q['to_node'].split('-')[0] == "producers"):

                raise ValueError(
                    ""
                    "Pipe id {} goes from producers to producers. "
                    "This is not allowed!".format(p))

            if ((q['from_node'].split('-')[0] == "producers") and (
                    q['to_node'].split('-')[0] == "consumers")) or ((
                        q['from_node'].split('-')[0] == "consumers") and (
                            q['to_node'].split('-')[0] == "producers")):

                raise ValueError(
                    ""
                    "Pipe id {} goes from producers directly "
                    "to consumers, or vice versa. This is not allowed!"
                    "".format(p))

            if (q['from_node'].split('-')[0] == "forks") and (
                    q['to_node'].split('-')[0] == "consumers"):

                cons_id = q['to_node'].split('-')[1]

                if cons_id not in ids_consumers:
                    raise ValueError(
                        ""
                        "The consumer of pipe id {} does not exist!".format(p))

        pipe_to_cons_ids = list(self.thermal_network.components['pipes']['to_node'].values)
        pipe_to_cons_ids = [x.split('-')[1] for x in pipe_to_cons_ids
                            if x.split('-')[0] == 'consumers']

        for id in list(self.thermal_network.components['consumers'].index):
            if id not in pipe_to_cons_ids:
                raise ValueError(
                    "The consumer id {} has no connection the the grid!".format(id))

        # Check 3
        # check if all components of network are connected
        self.thermal_network.nx_graph = self.thermal_network.to_nx_graph()
        g = nx.Graph(self.thermal_network.nx_graph)
        if not nx.is_connected(g):
            nx_sum = [len(c) for c in sorted(nx.connected_components(g), key=len, reverse=True)]
            nx_detail = sorted(nx.connected_components(g), key=len, reverse=True)
            raise ValueError(
                "Network not connected! There are {} parts, with the following number of nodes: \n"
                "{} \n"
                "These are the separated elements/networks: \n"
                "{}".format(len(nx_sum), nx_sum, nx_detail)
            )

    def remove_inactive(self):
        """
        If the attribute active is present in any of the components
        columns, or in any the investment options tables,
        all rows with active == 0 are deleted, and the column active
        is deleted.
        """
        def clean_df(df):
            if 'active' in df.columns:
                v_new = df[df['active'] == 1].copy()
                v_new.drop('active', axis=1, inplace=True)
                df = v_new
            return df

        for k, v in self.thermal_network.components.items():
            self.thermal_network.components[k] = clean_df(v)

        pipes = self.invest_options['network']['pipes']
        self.invest_options['network']['pipes'] = clean_df(pipes)

        for node_typ in ['consumers', 'producers']:
            for k, v in self.invest_options[node_typ].items():
                self.invest_options[node_typ][k] = clean_df(v)

    def prepare_heat_demand(self):
        """
        This method performs the pre-processing of the heat demand data, depending on
        the given optimisation settings.

        - If attribute 'P_heat_max' not given at the consumers, the maximum heat demand
          is calculated from the timeseries and added the consumers table.
        - If the optimisation setting 'heat_demand' == scalar, the number of time steps
          of the optimisation is set to 1, and the 'P_heat_max' values are copied to the
          consumers heat flow sequences (which is always the input for the optimisation model).
        - The consumers heat flow sequences are multiplied by the simultaneity factor.
        - Finally, a sufficient length of the heat demand timeseries is checked.

        Returns
        -------
        Updated `.network.components['consumers']` and
        `.network.sequences['consumers']['heat_flow']`
        """
        def check_len_timeseries():
            """
            Check, if given number of timesteps of optimization exceeds the length
            of the given heat demand timeseries.
            """
            if self.settings['num_ts'] > \
                    len(self.thermal_network.sequences['consumers']['heat_flow'].index):
                raise ValueError(
                    'The length of the heat demand timeseries is not sufficient '
                    'for the given number of {} timesteps.'.format(
                        self.settings['num_ts']))

        # prepare heat data, whether global simultanity or timeseries
        if 'P_heat_max' not in list(
                self.thermal_network.components['consumers'].columns):
            df_max = self.thermal_network.sequences['consumers']['heat_flow'].max(). \
                to_frame(name='P_heat_max')

            self.thermal_network.components['consumers'] = \
                pd.concat([self.thermal_network.components['consumers'], df_max],
                          axis=1, join='outer', sort=False)

        # check, which optimization type should be performed
        if self.settings['heat_demand'] == 'scalar':
            # just single timestep optimization, overwrite previous!
            self.settings['num_ts'] = 1

            # new approach
            p_max = self.thermal_network.components['consumers']['P_heat_max']
            df_ts = pd.DataFrame(data=[p_max.values],
                                 columns=list(p_max.index),
                                 index=pd.Index([0], name='timestep'))

            # heat load is maximum heat load
            self.thermal_network.sequences['consumers']['heat_flow'] = df_ts

        # apply global simultaneity for demand series
        self.thermal_network.sequences['consumers']['heat_flow'] = \
            self.thermal_network.sequences['consumers']['heat_flow'] * \
            self.settings['simultaneity']

        check_len_timeseries()

    def check_existing(self):
        """
        Checks if the attributes `existing` and `hp_type` are given in the `pipes` table.
        If not, the attribute is added, and set to `None` / 0.

        Checks for all existing pipes, if the heatpipe type is given in the pipe type table
        `.invest_options['network']['pipes']`, and if the capacity is greater than zero.
        """

        # check whether there the 'existing' attribute is present at the pipes
        if 'existing' not in self.thermal_network.components['pipes'].columns:
            self.thermal_network.components['pipes']['existing'] = 0

        # create pipes attribute hp_type, if not in the table so far
        if 'hp_type' not in list(self.thermal_network.components['pipes'].columns):
            self.thermal_network.components['pipes']['hp_type'] = None

        edges = self.thermal_network.components['pipes']
        pipe_types = self.invest_options['network']['pipes']

        hp_list = list({x for x in edges['hp_type'].tolist()
                        if isinstance(x, str)})

        for hp in hp_list:
            if hp not in list(pipe_types['label_3']):
                raise ValueError(
                    "Existing heatpipe type {} is not in the list of "
                    "ACTIVE heatpipe investment options!".format(hp)
                )

        for r, c in edges[edges['existing'] == 1].iterrows():
            if c['capacity'] <= 0:
                raise ValueError(
                    "The `capacity` of the existing pipe with id {} must be greater than 0!"
                    "".format(r)
                )

    def setup_oemof_es(self):
        """The oemof solph energy system is initialised based on the settings,
         and filled with oemof-solph object:

         The oemof-solph objects of the *consumers* and *producers* are defined at the consumers
         and producers investment options.

         For the heating infrastructure, there is a *oemof.solph.Bus* added for every fork,
         and a pipe component for every pipe as defined in */network/pipes.csv*.
         """

        date_time_index = pd.date_range(self.settings['start_date'],
                                        periods=self.settings['num_ts'],
                                        freq=self.settings['frequence'])

        logger.define_logging(screen_level=logging.INFO)

        logging.info('Initialize the energy system')

        self.es = solph.EnergySystem(timeindex=date_time_index)

        logging.info('Create oemof objects')

        # add houses and generation
        for typ in ['consumers', 'producers']:
            self.nodes, self.buses = add_nodes_houses(
                self, self.nodes, self.buses, typ)

        logging.info('Producers, Consumers Nodes appended.')

        # add heating infrastructure
        self.nodes, self.buses = add_nodes_dhs(self, self.settings, self.nodes,
                                               self.buses)
        logging.info('DHS Nodes appended.')

        # add nodes and flows to energy system
        self.es.add(*self.nodes)

        logging.info('Energysystem has been created')

        if self.settings['print_logging_info']:
            print("*********************************************************")
            print("The following objects have been created:")
            for n in self.es.nodes:
                oobj = \
                    str(type(n)).replace("<class 'oemof.solph.", "").replace("'>",
                                                                             "")
                print(oobj + ':', n.label)
            print("*********************************************************")

    def setup(self):
        """
        Calls *remove_inactive()* *check_input()*, *prepare_heat_demand()*,
        *complete_exist_data()*, and *setup_oemof_es()*.
        """

        # removes all rows with attribute active == 0 - if 'active given
        self.remove_inactive()

        # initial check of pipes connections
        self.check_input()

        # pre-processes the heat demand data depending on optimisation settings
        self.prepare_heat_demand()

        # check if existing pipes are given
        self.check_existing()

        # set up oemof energy system
        self.setup_oemof_es()

    def solve(self):
        """Builds the oemof.solph.Model of the energysystem *es*."""

        logging.info('Build the operational model')
        self.om = solph.Model(self.es)

        if self.settings['solve_kw'] is None:
            s_kw = {"tee": True}
        else:
            s_kw = self.settings['solve_kw']

        logging.info('Solve the optimization problem')
        self.om.solve(
            solver=self.settings['solver'],
            solve_kwargs=s_kw,
            cmdline_options=self.settings.get('solver_cmdline_options', {}))

        if self.settings['write_lp_file']:
            filename = os.path.join(
                helpers.extend_basic_path('lp_files'), 'DHNx.lp')
            logging.info('Store lp-file in %s', filename)
            self.om.write(filename, io_options={'symbolic_solver_labels': True})

        self.es.results['main'] = solph.processing.results(self.om)
        self.es.results['meta'] = solph.processing.meta_results(self.om)

    def get_results_edges(self):
        """Postprocessing of the investment results of the pipes."""

        def get_invest_val(lab):

            res = self.es.results['main']

            outflow = [x for x in res.keys()
                       if x[1] is not None
                       if lab == str(x[0].label)]

            if len(outflow) > 1:
                print('Multiple IDs!')

            try:
                invest = res[outflow[0]]['scalars']['invest']
            except (KeyError, IndexError):
                try:
                    # that's in case of a one timestep optimisation due to
                    # an oemof bug in outputlib
                    invest = res[outflow[0]]['sequences']['invest'][0]
                except (KeyError, IndexError):
                    # this is in case there is no bi-directional heatpipe, e.g. at
                    # forks-consumers, producers-forks
                    invest = 0

            # the rounding is performed due to numerical issues
            return round(invest, 6)

        def get_invest_status(lab):

            res = self.es.results['main']

            outflow = [x for x in res.keys()
                       if x[1] is not None
                       if lab == str(x[0].label)]

            try:
                invest_status = res[outflow[0]]['scalars']['invest_status']
            except (KeyError, IndexError):
                try:
                    # that's in case of a one timestep optimisation due to
                    # an oemof bug in outputlib
                    invest_status = res[outflow[0]]['sequences']['invest_status'][0]
                except (KeyError, IndexError):
                    # this is in case there is no bi-directional heatpipe, e.g. at
                    # forks-consumers, producers-forks
                    invest_status = 0

            return invest_status

        def get_hp_results(p):

            hp_lab = p['label_3']
            label_base = 'infrastructure_' + 'heat_' + hp_lab + '_'

            # maybe slow approach with lambda function
            df[hp_lab + '.' + 'dir-1'] = df['from_node'] + '-' + df['to_node']
            df[hp_lab + '.' + 'size-1'] = df[hp_lab + '.' + 'dir-1'].apply(
                lambda x: get_invest_val(label_base + x))
            df[hp_lab + '.' + 'dir-2'] = df['to_node'] + '-' + df['from_node']
            df[hp_lab + '.' + 'size-2'] = df[hp_lab + '.' + 'dir-2'].apply(
                lambda x: get_invest_val(label_base + x))

            df[hp_lab + '.' + 'size'] = \
                df[[hp_lab + '.' + 'size-1', hp_lab + '.' + 'size-2']].max(axis=1)

            # get direction of pipes
            for r, c in df.iterrows():
                if c[hp_lab + '.' + 'size-1'] > c[hp_lab + '.' + 'size-2']:
                    df.at[r, hp_lab + '.direction'] = 1
                elif c[hp_lab + '.' + 'size-1'] < c[hp_lab + '.' + 'size-2']:
                    df.at[r, hp_lab + '.direction'] = -1
                else:
                    df.at[r, hp_lab + '.direction'] = 0

            if p['nonconvex']:
                df[hp_lab + '.' + 'status-1'] = df[hp_lab + '.' + 'dir-1'].apply(
                    lambda x: get_invest_status(label_base + x))
                df[hp_lab + '.' + 'status-2'] = df[hp_lab + '.' + 'dir-2'].apply(
                    lambda x: get_invest_status(label_base + x))
                df[hp_lab + '.' + 'status'] = \
                    df[[hp_lab + '.' + 'status-1', hp_lab + '.' + 'status-2']].max(axis=1)

                for r, c in df.iterrows():
                    if df.at[r, hp_lab + '.' + 'status-1'] + \
                            df.at[r, hp_lab + '.' + 'status-2'] > 1:
                        print(
                            "Investment status of pipe id {} is 1 for both dircetions!"
                            " This is not allowed!".format(r)
                        )
                    if (df.at[r, hp_lab + '.' + 'status-1'] == 1 and df.at[
                            r, hp_lab + '.' + 'size-1'] == 0) \
                            or\
                            (df.at[r, hp_lab + '.' + 'status-2'] == 1 and df.at[
                                r, hp_lab + '.' + 'size-2'] == 0):
                        print(
                            "Investment status of pipe id {} is 1, and capacity is 0!"
                            "What happend?!".format(r)
                        )

            return df

        def check_multi_dir_invest(hp_lab):

            df_double_invest = \
                df[(df[hp_lab + '.' + 'size-1'] > 0) & (df[hp_lab + '.' + 'size-2'] > 0)]

            if self.settings['print_logging_info']:
                print('***')
                if df_double_invest.empty:
                    print('There is NO investment in both directions at the'
                          'following pipes for "', hp_lab, '"')
                else:
                    print('There is an investment in both directions at the'
                          'following pipes for "', hp_lab, '":')
                    print('----------')
                    print(' id | from_node | to_node | size-1 | size-2 ')
                    print('============================================')
                    for r, c in df_double_invest.iterrows():
                        print(r, ' | ', c['from_node'], ' | ', c['to_node'],
                              ' | ', c[hp_lab + '.' + 'size-1'], ' | ',
                              c[hp_lab + '.' + 'size-2'], ' | ')
                    print('----------')

        def catch_up_results():

            def check_invest_label(hp_type, edge_id):
                """
                If there is already a heatpipe type, an error is raised,
                because only one investment type for each edges makes sense.
                """
                if isinstance(hp_type, str):
                    raise ValueError(
                        "Pipe id {} already has an investment > 0!".format(edge_id))

            df['hp_type'] = None
            df['capacity'] = float(0)
            df['direction'] = 0

            for ahp in active_hp:
                # p = df_hp[df_hp['label_3'] == ahp].squeeze()   # series of heatpipe
                for r, c in df.iterrows():
                    if c[ahp + '.size'] > 0:
                        check_invest_label(c['hp_type'], id)
                        df.at[r, 'hp_type'] = ahp
                        df.at[r, 'capacity'] = c[ahp + '.size']
                        df.at[r, 'direction'] = c[ahp + '.direction']

        def recalc_costs_losses():

            df['costs'] = float(0)
            df['losses'] = float(0)

            for r, c in df.iterrows():
                if c['capacity'] > 0:
                    hp_lab = c['hp_type']
                    # select row from heatpipe type table
                    hp_p = df_hp[df_hp['label_3'] == hp_lab].squeeze()
                    if hp_p['nonconvex'] == 1:
                        df.at[r, 'costs'] = c['length'] * (
                            c['capacity'] * hp_p['capex_pipes'] + hp_p['fix_costs']
                        )
                        df.at[r, 'losses'] = c['length'] * (
                            c['capacity'] * hp_p['l_factor'] + hp_p['l_factor_fix']
                        )
                    elif hp_p['nonconvex'] == 0:
                        df.at[r, 'costs'] = c['length'] * c['capacity'] * hp_p['capex_pipes']
                        # Note, that a constant loss is possible also for convex
                        df.at[r, 'losses'] = c['length'] * (
                            c['capacity'] * hp_p['l_factor'] + hp_p['l_factor_fix']
                        )

        # use pipes dataframe as base and add results as new columns to it
        df = self.thermal_network.components['pipes']

        # only select not existing pipes
        df = df[df['existing'] == 0].copy()

        # remove input data
        df = df[['from_node', 'to_node', 'length']].copy()

        # putting the results of the investments in heatpipes to the pipes:
        df_hp = self.invest_options['network']['pipes']

        # list of active heat pipes
        active_hp = list(df_hp['label_3'].values)

        for hp in active_hp:
            hp_param = df_hp[df_hp['label_3'] == hp].squeeze()
            get_hp_results(hp_param)
            check_multi_dir_invest(hp)

        catch_up_results()

        recalc_costs_losses()

        return df[['from_node', 'to_node', 'length', 'hp_type', 'capacity', 'direction',
                   'costs', 'losses']]


def optimize_operation(thermal_network):
    r"""
    Takes a thermal network and returns the result of
    the operational optimization.
    """
    model = OemofOperationOptimizationModel(thermal_network)

    model.solve()

    results = model.get_results()

    return results


def setup_optimise_investment(
        thermal_network, invest_options, heat_demand='scalar', num_ts=1,
        time_res=1, start_date='1/1/2018', frequence='H', solver='cbc',
        solve_kw=None, solver_cmdline_options=None, simultaneity=1,
        bidirectional_pipes=False, dump_path=None, dump_name='dump.oemof',
        print_logging_info=False, write_lp_file=False):
    """
    Function for setting up the oemof solph operational Model.

    Parameters
    ----------
    thermal_network : ThermalNetwork
        See the ThermalNetwork class.
    invest_options : dict
        Dictionary holding the investment options for the district heating system.
    heat_demand : str
        'scalar': Peak heat load is used as heat consumers’ heat demand.
        'series': Heat load time-series is used.
    num_ts : int
        Number of time steps of optimisation.
    time_res : float
        Time resolution.
    start_date : str or datetime-like
        Startdate for oemof optimisation.
    frequence : str or DateOffset
        Lenght of period.
    solver : str
        Name of solver.
    solve_kw : dict
        Solver kwargs.
    solver_cmdline_options : dict
        Dictionary with command line options for solver.
    simultaneity : float
        Simultaneity factor.
    bidirectional_pipes : bool
        Bidirectional pipes leads to bi-directional flow attributes at the heatpipeline components
        {‘min’: -1, bidirectional: True}.
    dump_path : str
        If a dump path is provided, the oemof dump file is stored.
    dump_name : str
        Name of dump file.
    print_logging_info : bool
        Additional logging info is printed.
    write_lp_file : bool
        Linear program file is stored (‘User/.oemof/lp_files/DHNx.lp’).

    Returns
    -------
    oemof.solph.Model : The oemof.solph.Model is build.
    """
    if heat_demand not in ['scalar', 'series']:
        raise ValueError(
            'The settings attribute *heat_demand* must be "scalar" or "series"!'
        )

    if solver_cmdline_options is None:
        solver_cmdline_options = {}

    settings = {
        'heat_demand': heat_demand,
        'num_ts': num_ts,
        'time_res': time_res,
        'start_date': start_date,
        'frequence': frequence,
        'solver': solver,
        'solve_kw': solve_kw,
        'solver_cmdline_options': solver_cmdline_options,
        'simultaneity': simultaneity,
        'bidirectional_pipes': bidirectional_pipes,
        'dump_path': dump_path,
        'dump_name': dump_name,
        'print_logging_info': print_logging_info,
        'write_lp_file': write_lp_file,
    }

    model = OemofInvestOptimizationModel(thermal_network, settings, invest_options)

    return model


def solve_optimisation_investment(model):
    """

    Parameters
    ----------
    model : oemof.solph.Model
        The oemof model, which is optimized.

    Returns
    -------
    dict : Results of optimisation. Contains:
        - 'oemof' : Complete "oemof" results of the energy system optimisation (.results['main']).
        - 'oemof_meta' : Meta results of oemof solph optimisation.
        - 'components' : 'pipes' : Investment results of pipes.
    """

    model.solve()

    if model.settings['dump_path'] is not None:
        my_es = model.es
        my_es.dump(dpath=model.settings['dump_path'], filename=model.settings['dump_name'])
        print('oemof Energysystem stored in "{}"'.format(model.settings['dump_path']))

    edges_results = model.get_results_edges()

    results = {'oemof': model.es.results['main'],
               'oemof_meta': model.es.results['meta'],
               'components': {'pipes': edges_results}}

    return results



__version__ = "0.0.1"

from . import input_output
from . import network
from . import plotting
from . import model
from . import simulation
from . import dhn_from_osm
from . import optimization
from . import optimization_add_components
from . import optimization_dhs_nodes
from . import optimization_oemof_heatpipe



# -*- coding: utf-8

"""
This module is designed to hold functions to convert ThermalNetworks into
networkx a graph.

This file is part of project dhnx (). It's copyrighted
by the contributors recorded in the version control history of the file,
available from its original location:

SPDX-License-Identifier: MIT
"""

import pandas as pd
import networkx as nx


def thermal_network_to_nx_graph(thermal_network):
    r"""

    Parameters
    ----------
    thermal_network : dhnx.network.ThermalNetwork

    Returns
    -------
    nx_graph : nx:MultiDigraph
        networkx graph of thermal_network
    """
    nx_graph = nx.DiGraph()  # TODO: Check if this line can be removed.

    edges = thermal_network.components['pipes'].copy()

    edge_attr = list(edges.columns)

    edge_attr.remove('from_node')

    edge_attr.remove('to_node')

    nx_graph = nx.from_pandas_edgelist(
        edges,
        'from_node',
        'to_node',
        edge_attr=edge_attr,
        create_using=nx_graph
    )

    nodes = {
        list_name: thermal_network.components[list_name].copy() for list_name in [
            'consumers',  # TODO: Do not hard code these here
            'producers',
            'forks'
        ]
    }

    for k, v in nodes.items():
        v.index = [k + '-' + str(id) for id in v.index]

    nodes = pd.concat(nodes.values(), sort=True)

    node_attrs = {node_id: dict(data) for node_id, data in nodes.iterrows()}

    nx.set_node_attributes(nx_graph, node_attrs)

    return nx_graph


def nx_graph_to_thermal_network(nx_graph):
    r"""
    Creates ThermalNetwork from nx.MultiDigraph

    Parameters
    ----------
    nx_graph : nx.MultiDigraph

    Returns
    -------
    thermal_network : ThermalNetwork
    """
    raise NotImplementedError('This feature is not implemented yet.')


def write_edge_data_to_graph(series, graph_in, var_name=None):
    r"""
    Writes data describing the edges to the graph. Data has to
    be a pd.Series labeled with the (from, to). If the series has
    a name, the data will be stored in the graph under that name.
    If not, `var_name` has to be provided.

    Parameters
    ----------
    series
    graph_in
    var_name

    Returns
    -------

    """
    graph = graph_in.copy()

    assert isinstance(series, pd.Series), \
        "Have to pass a pandas Series."

    if var_name:
        pass
    elif series.name:
        var_name = series.name
    else:
        raise ValueError(r"Have to either pass Series with name or provide var_name.")

    for index, value in series.iteritems():

        graph.edges[index][var_name] = value

    return graph



# -*- coding: utf-8

"""
This module is designed to hold importers and exporters to different file formats.

This file is part of project dhnx (). It's copyrighted
by the contributors recorded in the version control history of the file,
available from its original location:

SPDX-License-Identifier: MIT
"""

import logging
import os

from addict import Dict
import pandas as pd
import networkx as nx
import numpy as np

try:
    from shapely.geometry import Point
    from shapely.geometry import LineString

except ImportError:
    print("Need to install shapely to download from osm.")

try:
    import geopandas as gpd

except ImportError:
    print("Need to install geopandas to download from osm.")

try:
    import osmnx as ox

except ImportError:
    print("Need to install osmnx to download from osm.")

from dhnx.dhn_from_osm import connect_points_to_network


logger = logging.getLogger()


class NetworkImporter():
    r"""
    Generic Importer object for network.ThermalNetwork
    """
    def __init__(self, thermal_network, basedir):
        thermal_network.is_consistent()
        self.thermal_network = thermal_network
        self.basedir = basedir

    def load(self):
        pass


class NetworkExporter():
    r"""
    Generic Exporter object for network.ThermalNetwork
    """
    def __init__(self, thermal_network, basedir):
        thermal_network.is_consistent()
        self.thermal_network = thermal_network
        self.basedir = basedir

    def save(self):
        pass


class CSVNetworkImporter(NetworkImporter):
    r"""
    Imports thermal networks from directory with csv-files.
    """
    def load_component_table(self, list_name):

        if list_name not in self.thermal_network.available_components.list_name.values:
            raise KeyError(f"Component '{list_name}'is not "
                           f"part of the available components.")

        file_name = list_name + '.csv'
        component_table = pd.read_csv(os.path.join(self.basedir, file_name), index_col=0)

        return component_table

    def load_sequence(self, list_name, attr_name):

        if list_name not in self.thermal_network.available_components.list_name.values:
            raise KeyError(f"Component '{list_name}' is not "
                           f"part of the available components.")

        if attr_name not in self.thermal_network.component_attrs[list_name]:
            logger.info("Attribute '%s' is not "
                        "part of the component attributes.", attr_name)

        file_name = '-'.join([list_name, attr_name]) + '.csv'
        sequence = pd.read_csv(os.path.join(self.basedir, 'sequences', file_name), index_col=0)

        return sequence

    def load(self):
        for name in os.listdir(self.basedir):

            if name.endswith('.csv'):

                list_name = os.path.splitext(name)[0]

                self.thermal_network.components[list_name] = self.load_component_table(list_name)

                self.thermal_network.set_defaults()

            elif os.path.isdir(os.path.join(self.basedir, name)):

                assert name in ['sequences'], f"Unknown directory name. Directory '{name}' " \
                                              f"is not a defined subdirectory."

                for sequence_name in os.listdir(os.path.join(self.basedir, name)):

                    assert sequence_name.endswith('.csv'), f"Inappropriate filetype of '{name}'" \
                                                           f"for csv import."

                    list_name, attr_name = tuple(sequence_name.split('-'))

                    attr_name = os.path.splitext(attr_name)[0]

                    if list_name not in self.thermal_network.sequences:
                        self.thermal_network.sequences[list_name] = Dict()

                    self.thermal_network.sequences[list_name][attr_name] =\
                        self.load_sequence(list_name, attr_name)

            else:
                raise ImportError(f"Inappropriate filetype of '{name}' for csv import.")

        self.thermal_network.is_consistent()

        self.thermal_network.set_timeindex()

        return self.thermal_network


class CSVNetworkExporter(NetworkExporter):
    r"""
    Exports thermal networks to directory with csv-files.
    """
    def __init__(self, thermal_network, basedir):
        super().__init__(thermal_network, basedir)
        if not os.path.exists(self.basedir):
            os.mkdir(self.basedir)

    def save_component_table(self, component_table, name):
        component_table.to_csv(os.path.join(self.basedir, name))

    def save_sequence(self, list_name, attr_name, sequence):

        sequence_dir = os.path.join(self.basedir, 'sequences')

        if not os.path.exists(sequence_dir):

            os.mkdir(sequence_dir)

        file_name = '-'.join([list_name, attr_name]) + '.csv'

        sequence.to_csv(os.path.join(self.basedir, 'sequences', file_name))

    def save(self):
        for list_name, component_table in self.thermal_network.components.items():
            if not component_table.empty:
                filename = list_name + '.csv'
                self.save_component_table(component_table, filename)

        for list_name, subdict in self.thermal_network.sequences.items():

            for attr_name, sequence in subdict.items():

                self.save_sequence(list_name, attr_name, sequence)


class OSMNetworkImporter(NetworkImporter):
    r"""
    Imports thermal networks from OSM data.

    Not yet implemented.
    """
    def __init__(self, thermal_network, place, distance):
        super().__init__(thermal_network, None)

        self.place = place

        self.distance = distance

    def download_street_network(self):

        print('Downloading street network...')

        graph = ox.graph_from_point(
            center_point=self.place, distance=self.distance
        )

        graph = ox.project_graph(graph)

        return graph

    def download_footprints(self):

        print('Downloading footprints...')

        footprints = ox.footprints.footprints_from_point(
            point=self.place, distance=self.distance
        )

        footprints = footprints.drop(labels='nodes', axis=1)

        footprints = ox.project_gdf(footprints)

        return footprints

    @staticmethod
    def remove_duplicate_edges(graph):

        graph = nx.DiGraph(graph)
        return graph

    @staticmethod
    def remove_self_loops(graph):

        self_loop_edges = list(nx.selfloop_edges(graph))

        graph.remove_edges_from(self_loop_edges)

        return graph

    @staticmethod
    def graph_to_gdfs(G, nodes=True, edges=True, node_geometry=True, fill_edge_geometry=True):
        """
        Convert a graph into node and/or edge GeoDataFrames. Adapted from osmnx for DiGraph.

        Parameters
        ----------
        G : networkx digraph
        nodes : bool
            if True, convert graph nodes to a GeoDataFrame and return it
        edges : bool
            if True, convert graph edges to a GeoDataFrame and return it
        node_geometry : bool
            if True, create a geometry column from node x and y data
        fill_edge_geometry : bool
            if True, fill in missing edge geometry fields using origin and
            destination nodes

        Returns
        -------
        GeoDataFrame or tuple
            gdf_nodes or gdf_edges or both as a tuple
        """
        if not (nodes or edges):
            raise ValueError('You must request nodes or edges, or both.')

        to_return = []

        if nodes:

            nodes, data = zip(*G.nodes(data=True))
            gdf_nodes = gpd.GeoDataFrame(list(data), index=nodes)
            if node_geometry:
                gdf_nodes['geometry'] = gdf_nodes.apply(lambda row: Point(row['x'], row['y']),
                                                        axis=1)
                gdf_nodes.set_geometry('geometry', inplace=True)
            gdf_nodes.crs = G.graph['crs']
            gdf_nodes.gdf_name = '{}_nodes'.format(G.graph['name'])

            to_return.append(gdf_nodes)

        if edges:

            # create a list to hold our edges, then loop through each edge in the
            # graph
            edges = []
            for u, v, data in G.edges(data=True):

                # for each edge, add key and all attributes in data dict to the
                # edge_details
                edge_details = {'u': u, 'v': v}
                for attr_key in data:
                    edge_details[attr_key] = data[attr_key]

                # if edge doesn't already have a geometry attribute, create one now
                # if fill_edge_geometry==True
                if 'geometry' not in data:
                    if fill_edge_geometry:
                        point_u = Point((G.nodes[u]['x'], G.nodes[u]['y']))
                        point_v = Point((G.nodes[v]['x'], G.nodes[v]['y']))
                        edge_details['geometry'] = LineString([point_u, point_v])
                    else:
                        edge_details['geometry'] = np.nan

                edges.append(edge_details)

            # create a GeoDataFrame from the list of edges and set the CRS
            gdf_edges = gpd.GeoDataFrame(edges)
            gdf_edges.crs = G.graph['crs']
            gdf_edges.gdf_name = '{}_edges'.format(G.graph['name'])

            to_return.append(gdf_edges)

        if len(to_return) > 1:
            return tuple(to_return)

        return to_return[0]

    @staticmethod
    def get_building_midpoints(footprints):

        building_midpoints = gpd.GeoDataFrame(
            footprints.geometry.centroid,
            columns=['geometry'],
            crs=footprints.geometry.centroid.crs
        )

        return building_midpoints

    def graph_to_component_dfs(self, graph, building_midpoints):

        # get nodes and edges from graph
        nodes, edges = self.graph_to_gdfs(graph)

        nodes = nodes.loc[:, ['geometry']]

        edges = edges.loc[:, ['u', 'v', 'geometry']]

        nodes = nodes.to_crs("epsg:4326")

        edges = edges.to_crs("epsg:4326")

        building_midpoints = building_midpoints.to_crs("epsg:4326")

        endpoints, forks, pipes = connect_points_to_network(
            building_midpoints, nodes, edges)

        pipes = pipes.rename(columns={'u': 'from_node', 'v': 'to_node'})

        forks['component_type'] = 'Fork'

        consumers = endpoints

        consumers['component_type'] = 'Consumer'

        # Update names of nodes in pipe's from_node/to_node
        rename_nodes = {i: 'forks-' + str(i) for i in forks.index}

        rename_nodes.update({i: 'consumers-' + str(i) for i in consumers.index})

        pipes['from_node'].replace(rename_nodes, inplace=True)

        pipes['to_node'].replace(rename_nodes, inplace=True)

        pipes['length'] = pipes['geometry'].length

        for node in [consumers, forks]:
            node['lat'] = node.geometry.y
            node['lon'] = node.geometry.x

        component_dfs = {
            'consumers': consumers,
            'forks': forks,
            'pipes': pipes
        }

        for df in component_dfs.values():
            df.index.name = 'id'

        return component_dfs

    def add_component_data_to_network(self, components):

        for k, v in components.items():
            self.thermal_network.components[k] = v

        self.thermal_network.is_consistent()

    def process(self, graph, footprints):

        print('Processing...')

        graph = self.remove_self_loops(graph)

        graph = self.remove_duplicate_edges(graph)

        graph = nx.relabel.convert_node_labels_to_integers(graph)

        building_midpoints = self.get_building_midpoints(footprints)

        component_dfs = self.graph_to_component_dfs(graph, building_midpoints)

        return component_dfs

    def load(self):

        graph = self.download_street_network()

        footprints = self.download_footprints()

        footprints = ox.project_gdf(footprints)

        component_dfs = self.process(graph, footprints)

        self.add_component_data_to_network(component_dfs)

        return self.thermal_network


class GDFNetworkExporter(NetworkExporter):
    r"""
    Exports thermal networks to geopandas.GeoDataFrame.

    Not yet implemented.
    """


def load_component_attrs(dir_name, available_components):

    component_attrs = {}

    for file_name in os.listdir(dir_name):

        list_name = os.path.splitext(file_name)[0]

        assert list_name in available_components.list_name.values, f"Unknown component {list_name}"\
                                                                   " not in available components."

        df = pd.read_csv(os.path.join(dir_name, file_name), index_col=0)

        component_attrs[list_name] = df.T.to_dict()

    return Dict(component_attrs)


def load_invest_options(path):
    """
    This function loads the tables for the investment options.

    The structure must like this:

    .. _folder_structure_invest:

    .. code-block:: txt

        tree
        ├── network
        |   └── pipes.csv           # (required)
        |
        ├── consumers
        |   ├── bus.csv             # (required)
        |   ├── demand.csv          # (required)
        |   ├── source.csv          # (optional)
        |   ├── storages.csv        # (optional)
        |   └── transformer.csv     # (optional)
        |
        └── producers
            ├── bus.csv             # (required)
            ├── demand.csv          # (optional)
            ├── source.csv          # (required)
            ├── storages.csv        # (optional)
            └── transformer.csv     # (optional)

    Parameters
    ----------
    path : str
        Directory path to the folder with investment data.

    Returns
    -------
    dict : A dictionary with the dataframes of the given csv files.

    """

    dict = {}

    for name in os.listdir(path):

        dict_sub = {}

        for sub in os.listdir(os.path.join(path, name)):
            key = sub.split('.csv')[0]
            val = pd.read_csv(os.path.join(path, name, sub))
            dict_sub.update([(key, val)])

        dict.update([(name, dict_sub)])

    return dict


def save_results(results, results_dir):

    if not os.path.exists(results_dir):
        os.mkdir(results_dir)

    for k, v in results.items():
        if v is not None:
            v.to_csv(os.path.join(results_dir, k + '.csv'), header=True)



import addict


class Dict(addict.Dict):
    def __init__(self, *args, **kwargs):
        super().__init__(self, *args, **kwargs)

    def __repr__(self):
        overview = ['* ' + str(key) for key, value in self.items()]
        return '\n'.join(overview)


def sum_ignore_none(*items):
    not_none = [value for value in items if value is not None]

    if not_none:
        sum_ignoring_none = sum(not_none)

    else:
        sum_ignoring_none = None

    return sum_ignoring_none



# -*- coding: utf-8

"""
This module is designed to hold functions for importing networks and building
footprints from openstreetmap.

This file is part of project dhnx (). It's copyrighted
by the contributors recorded in the version control history of the file,
available from its original location:

SPDX-License-Identifier: MIT
"""

try:
    import geopandas as gpd

except ImportError:
    print("Need to install geopandas to process osm data.")

import pandas as pd

try:
    from shapely.ops import nearest_points
    from shapely.geometry import LineString

except ImportError:
    print("Need to install shapely to download from osm.")


def connect_points_to_network(points, nodes, edges):
    r"""

    Parameter
    ---------
    points :  geopandas.GeoDataFrame
        Points to connect to the network

    nodes : geopandas.GeoDataFrame
        Nodes of the network

    edges : geopandas.GeoDataFrame
        Edges of the network

    Returns
    -------
    points : geopandas.GeoDataFrame
        Points connected to the network

    nodes : geopandas.GeoDataFrame
        Original nodes of the network and
        nearest connecting points on the
        network's edges.

    edges :  geopandas.GeoDataFrame
        Edges of the network.
    """
    edges_united = edges.unary_union

    len_nodes = len(nodes)
    len_points = len(points)

    # assign ids to new points
    n_points = []
    n_nearest_points = []
    n_edges = []

    for i, point in enumerate(points.geometry):
        id_nearest_point = len_nodes + i

        id_point = len_nodes + len_points + i

        nearest_point = nearest_points(edges_united, point)[0]

        n_points.append([id_point, point])

        n_nearest_points.append([id_nearest_point, nearest_point])

        n_edges.append([id_point, id_nearest_point, LineString([point, nearest_point])])

    n_points = gpd.GeoDataFrame(
        n_points,
        columns=['index', 'geometry']).set_index('index')

    n_nearest_points = gpd.GeoDataFrame(
        n_nearest_points,
        columns=['index', 'geometry']).set_index('index')

    n_edges = gpd.GeoDataFrame(n_edges, columns=['u', 'v', 'geometry'])

    joined_nodes = pd.concat([nodes, n_nearest_points], sort=True)
    joined_edges = pd.concat([edges, n_edges], sort=True)

    return n_points, joined_nodes, joined_edges



# -*- coding: utf-8

"""
This module is designed to hold optimization model implementations. The
implementation makes use of oemof-solph.

This file is part of project dhnx (). It's copyrighted
by the contributors recorded in the version control history of the file,
available from its original location:

SPDX-License-Identifier: MIT
"""


import oemof.solph as solph
import dhnx.optimization_oemof_heatpipe as oh
import dhnx.optimization_add_components as ac


def add_nodes_dhs(opti_network, gd, nodes, busd):
    """
    Based on the *forks* and *pipes* of the *ThermalNetwork* of the
    *OemofInvestOptimisationModel*, the oemof-solph components for the district heating
    network optimisation are initialised.

    For all *forks*, a `solph.Bus` with the label `infrastructure_heat_bus_forks-<id>`
    (string representation) is generated.

    For all *pipes*, a 'HeatPipeline' with the label
    `infrastructure_heat_<type>_<from_node>-<to_node>` (string representation)
    is generated. Depending on the attributes in *pipes*,
    an investment pipe, or an existing pipe is built.


    Parameters
    ----------
    opti_network : OemofInvestOptimisationModel
    gd : dict
        General optimisation settings.
    nodes : list
        List for collecting all oemof-solph objects.
    busd : dict
        Dictionary collecting all oemof-solph Buses. Keys: labels of the oemof-solph Buses;
        Values: oemof-solph Buses.

    Returns
    -------
    list, dict : List of all oemof-solph objects and dictionary of oemof-solph Buses.
    """

    d_labels = {}

    d_labels['l_2'] = 'heat'
    d_labels['l_3'] = 'bus'

    for n, _ in opti_network.thermal_network.components['forks'].iterrows():
        d_labels['l_4'] = 'forks-' + str(n)
        d_labels['l_1'] = 'infrastructure'
        l_bus = oh.Label(d_labels['l_1'], d_labels['l_2'], d_labels['l_3'],
                         d_labels['l_4'])
        bus = solph.Bus(label=l_bus)
        nodes.append(bus)
        busd[l_bus] = bus

    # add heatpipes for all lines
    for p, q in opti_network.thermal_network.components['pipes'].iterrows():

        pipe_data = opti_network.invest_options['network']['pipes']

        d_labels['l_1'] = 'infrastructure'
        d_labels['l_2'] = 'heat'

        if q['existing']:

            # terminate the first label
            l_1_in = 'infrastructure'
            l_1_out = 'infrastructure'

            typ_from = q['from_node'].split('-')[0]
            typ_to = q['to_node'].split('-')[0]

            if (typ_from == 'forks') and (typ_to == 'consumers'):
                l_1_out = 'consumers'

                start = q['from_node']
                end = q['to_node']
                b_in = busd[(l_1_in, d_labels['l_2'], 'bus', start)]
                b_out = busd[(l_1_out, d_labels['l_2'], 'bus', end)]
                d_labels['l_4'] = start + '-' + end
                nodes = ac.add_heatpipes_exist(pipe_data, d_labels, gd, q, b_in, b_out,
                                               nodes)

            elif (typ_from == 'consumers') and (typ_to == 'forks'):
                raise ValueError(
                    "Pipes must not go from 'consumers' to 'forks'!"
                    " Existing heatpipe id {}".format(p))

            elif (typ_from == 'forks') and (typ_to == 'producers'):
                l_1_in = 'producers'

                start = q['to_node']
                end = q['from_node']
                b_in = busd[(l_1_in, d_labels['l_2'], 'bus', start)]
                b_out = busd[(l_1_out, d_labels['l_2'], 'bus', end)]
                d_labels['l_4'] = start + '-' + end
                nodes = ac.add_heatpipes_exist(pipe_data, d_labels, gd, q,
                                               b_in, b_out,
                                               nodes)

            elif (typ_from == 'producers') and (typ_to == 'forks'):
                l_1_in = 'producers'

                start = q['from_node']
                end = q['to_node']
                b_in = busd[(l_1_in, d_labels['l_2'], 'bus', start)]
                b_out = busd[(l_1_out, d_labels['l_2'], 'bus', end)]
                d_labels['l_4'] = start + '-' + end
                nodes = ac.add_heatpipes_exist(pipe_data, d_labels, gd, q, b_in, b_out,
                                               nodes)

            elif (typ_from == 'forks') and (typ_to == 'forks'):

                start = q['from_node']
                end = q['to_node']
                b_in = busd[(l_1_in, d_labels['l_2'], 'bus', start)]
                b_out = busd[(l_1_out, d_labels['l_2'], 'bus', end)]
                d_labels['l_4'] = start + '-' + end
                nodes = ac.add_heatpipes_exist(pipe_data, d_labels, gd, q, b_in, b_out,
                                               nodes)

            else:
                raise ValueError("Something wrong!")

        else:   # calls Investment heatpipeline function
            # connection of houses
            if q['to_node'].split('-')[0] == "consumers":

                start = q['from_node']
                end = q['to_node']
                b_in = busd[(d_labels['l_1'], d_labels['l_2'], 'bus', start)]
                b_out = busd[('consumers', d_labels['l_2'], 'bus', end)]

                d_labels['l_4'] = start + '-' + end

                nodes = ac.add_heatpipes(
                    pipe_data, d_labels, gd, q, b_in, b_out,
                    nodes)

            elif q['from_node'].split('-')[0] == "consumers":
                raise ValueError(
                    "Pipes must not go from 'consumers'!"
                    " Existing heatpipe id {}".format(p))

            elif q['to_node'].split('-')[0] == "producers":

                start = q['to_node']
                end = q['from_node']
                b_in = busd[('producers', d_labels['l_2'], 'bus', start)]
                b_out = busd[(d_labels['l_1'], d_labels['l_2'], 'bus', end)]

                d_labels['l_4'] = start + '-' + end

                nodes = ac.add_heatpipes(
                    pipe_data, d_labels,
                    gd, q, b_in, b_out,
                    nodes)

            elif q['from_node'].split('-')[0] == "producers":

                start = q['from_node']
                end = q['to_node']
                b_in = busd[('producers', d_labels['l_2'], 'bus', start)]
                b_out = busd[(d_labels['l_1'], d_labels['l_2'], 'bus', end)]

                d_labels['l_4'] = start + '-' + end

                nodes = ac.add_heatpipes(
                    pipe_data, d_labels, gd, q, b_in, b_out,
                    nodes)

            elif (q['from_node'].split('-')[0] == 'forks') and (
                    q['to_node'].split('-')[0] == 'forks'):

                b_in = busd[(d_labels['l_1'], d_labels['l_2'], 'bus', q['from_node'])]
                b_out = busd[(d_labels['l_1'], d_labels['l_2'], 'bus', q['to_node'])]
                d_labels['l_4'] = q['from_node'] + '-' + q['to_node']

                nodes = ac.add_heatpipes(
                    pipe_data, d_labels, gd, q, b_in, b_out, nodes)

                if not gd['bidirectional_pipes']:
                    # the heatpipes from fork to fork need to be created in
                    # both directions in this case bidiretional = False
                    b_in = busd[(
                        d_labels['l_1'], d_labels['l_2'], 'bus', q['to_node'])]
                    b_out = busd[(d_labels['l_1'], d_labels['l_2'], 'bus', q['from_node'])]
                    d_labels['l_4'] = q['to_node'] + '-' + q['from_node']

                    nodes = ac.add_heatpipes(
                        pipe_data, d_labels, gd, q, b_in, b_out, nodes)

            else:
                raise ValueError("Something wrong!")

    return nodes, busd


def add_nodes_houses(opti_network, nodes, busd, label_1):
    """
    For each *consumers*/*producers* of the *ThermalNetwork* of the
    *OemofInvestOptimisationModel*, the oemof-solph components for the *consumers*/*producers*
    are initialised depending on the given tables of the ``invest_options`` of the
    *OemofInvestOptimisationModel*.

    The tables of `invest_options` provide the information and attributes for the oemof-solph
    components, which should be build at every *consumer*/*producer*.

    The minimum requirement is to provide table with at least one *heat Bus* for all consumers
    and all producers, a table with a *heat sink* as demand for all consumers, and
    a table with at least one *heat source* for all producers.

    Additionally, further tables with *Transformer*, *Storages*, and further *Sources* and
    *Sinks* can be added.

    For the attributes for each table, you need to provide, please see:

    .. py:module::`~dhnx.optimization_modules.add_components`

    Parameters
    ----------
    opti_network : OemofInvestOptimisationModel

    gd : dict
        General optimisation settings.
    nodes : list
        List for collecting all oemof-solph objects.
    busd : dict
        Dictionary collecting all oemof-solph Buses. Keys: labels of the oemof-solph Buses;
        Values: oemof-solph Buses.
    label_1 : str
        First tag of the label, which is either `producers` or `consumers`.

    Returns
    -------
    list, dict : List of all oemof-solph objects and dictionary of oemof-solph Buses.

    """

    gen_data = opti_network.invest_options[label_1]     # genic data for all houses
    series = opti_network.thermal_network.sequences[label_1]    # sequences
    d_labels = {}

    for r, c in opti_network.thermal_network.components[label_1].iterrows():

        # heat bus is always necessary
        d_labels['l_1'] = label_1
        d_labels['l_4'] = label_1 + '-' + str(r)

        # add buses first, because other classes need to have them already
        nodes, busd = ac.add_buses(gen_data['bus'],
                                   d_labels, nodes, busd)

        for key, item in gen_data.items():

            if key == 'source':
                nodes = ac.add_sources(opti_network, item, c, d_labels, nodes, busd)

            if key == 'demand':
                nodes = ac.add_demand(item, d_labels, series, nodes, busd)

            if key == 'transformer':
                nodes = ac.add_transformer(item, d_labels, nodes, busd)

            if key == 'storages':
                nodes = ac.add_storage(item, d_labels, nodes, busd)

    return nodes, busd



# -*- coding: utf-8

"""
This module is designed to base classes for optimization and simulation models.

This file is part of project dhnx (). It's copyrighted
by the contributors recorded in the version control history of the file,
available from its original location:

SPDX-License-Identifier: MIT
"""

from addict import Dict


class Model():
    r"""
    Abstract base class for different kind of models.
    """
    def __init__(self, thermal_network):
        self.thermal_network = thermal_network
        self.setup()
        self.results = Dict()

    def setup(self):
        pass

    def solve(self):
        pass

    def get_results(self):
        pass

    def is_consistent(self):
        pass


class OperationOptimizationModel(Model):
    r"""
    Abstract base class for operational optimization models.
    """
    def __init__(self, thermal_network):
        super().__init__(thermal_network)
        self.is_consistent()
        self.results = ['a', 'b']

    def is_consistent(self):
        # TODO.
        pass


class InvestOptimizationModel(Model):
    r"""
    Abstract base class for investment optimization models.
    """
    def __init__(self, thermal_network):
        super().__init__(thermal_network)
        self.is_consistent()
        self.results = ['a', 'b']

    def is_consistent(self):
        # TODO.
        pass


class SimulationModel(Model):
    r"""
    Abstract base class for simulation models.
    """
    def __init__(self, thermal_network):
        super().__init__(thermal_network)
        self.is_consistent()
        self.results = ['a', 'b']

    def is_consistent(self):
        # TODO.
        pass



# -*- coding: utf-8

"""
This module is designed to hold the definition of the central ThermalNetwork
object and its components.

This file is part of project dhnx (). It's copyrighted
by the contributors recorded in the version control history of the file,
available from its original location:

SPDX-License-Identifier: MIT
"""

import os

import pandas as pd

from .graph import thermal_network_to_nx_graph
from .optimization import optimize_operation, setup_optimise_investment, \
    solve_optimisation_investment
from .helpers import Dict
from .input_output import CSVNetworkImporter, CSVNetworkExporter, load_component_attrs
from .simulation import simulate

dir_name = os.path.dirname(__file__)

available_components = pd.read_csv(os.path.join(dir_name, 'components.csv'), index_col=0)

component_attrs = load_component_attrs(os.path.join(dir_name, 'component_attrs'),
                                       available_components)

required_attrs = {
    list_name: [
        attr for attr, specs in attrs.items() if specs.requirement == 'required'
    ] for list_name, attrs in component_attrs.items()
}

default_attrs = {
    list_name: {
        attr: specs.default for attr, specs in attrs.items()
        if not pd.isnull(specs.default)
    } for list_name, attrs in component_attrs.items()
}


class ThermalNetwork():
    r"""
    Class representing thermal (heating/cooling) networks.

    Parameters
    ----------
    availalable_components
    component_attrs
    components
    sequences
    results
    graph

    Examples
    --------
    >>> from dhnx.network import ThermalNetwork
    >>> tnw = ThermalNetwork('csv_folder')
    >>> tnw.is_consistent()
    True
    """
    def __init__(self, dirname=None):

        self.available_components = available_components
        self.component_attrs = component_attrs
        self.components = Dict({key: pd.DataFrame() for key in available_components.list_name})
        self.sequences = Dict()
        self.results = Dict()
        self.timeindex = None
        self.graph = None

        if dirname is not None:
            try:
                self.from_csv_folder(dirname)

            except ImportError:
                print('Failed to import file.')

    def __repr__(self):
        r"""
        This method defines what is returned if you perform print() or str()
        on a ThermalNetwork.
        """
        summary = ''
        for component, data in self.components.items():
            count = len(data)
            if count > 0:
                summary += ' * ' + str(count) + ' ' + component + '\n'

        if summary == '':
            return "Empty dhnx.network.ThermalNetwork object containing no components."

        return f"dhnx.network.ThermalNetwork object with these components\n{summary}"

    def from_csv_folder(self, dirname):
        importer = CSVNetworkImporter(self, dirname)

        self = importer.load()

        return self

    def to_csv_folder(self, dirname):
        exporter = CSVNetworkExporter(self, dirname)

        exporter.save()

    def to_nx_graph(self):
        nx_graph = thermal_network_to_nx_graph(self)

        return nx_graph

    def set_defaults(self):
        r"""
        Sets default values on component DataFrames.

        Returns
        -------
        None
        """

        for component, data in self.components.items():
            for default_name, default_value in default_attrs[component].items():

                data[default_name] = default_value

    def add(self, class_name, id, **kwargs):
        r"""
        Adds a row with id to the component DataFrame specified by class_name.

        Parameters
        ----------
        class_name
        id
        kwargs
        """
        assert class_name in available_components.index,\
            f"Component class {class_name} is not within the available components" \
            f" {available_components.index}."

        list_name = available_components.loc[class_name].list_name

        assert id not in self.components[list_name].index,\
            f"There is already a component with the id {id}."

        # check if required parameters are in kwargs
        missing_required = list(set(required_attrs[list_name]) - kwargs.keys())

        missing_required.remove('id')

        if bool(missing_required):
            raise ValueError(f"Required attributes {missing_required} are not given")

        # if not defined in kwargs, set default attributes
        component_data = default_attrs[list_name].copy()

        component_data.update(kwargs)

        # add to component DataFrame
        for key, value in component_data.items():
            self.components[list_name].loc[id, key] = value

    def remove(self, class_name, id):
        r"""
        Removes the row with id from the component DataFrame specified by class_name.

        Parameters
        ----------
        class_name : str
            Name of the component class
        id : int
            id of the component to remove
        """
        assert class_name in available_components.index,\
            "Component class '{}' is not within the available_components."

        list_name = available_components.loc[class_name].list_name

        assert id in self.components[list_name].index,\
            f"There is no component with the id {id}."

        self.components[list_name].drop(id, inplace=True)

    def is_consistent(self):
        r"""
        Checks that
         * pipes connect to existing nodes,
         * pipes do not connect a node with itself,
         * there are no duplicate pipes between two nodes.
        """
        nodes = {list_name: self.components[list_name].copy() for list_name in [
            'consumers',
            'producers',
            'forks'
        ]}

        for k, v in nodes.items():
            v.index = [k + '-' + str(id) for id in v.index]

        nodes = pd.concat(nodes.values(), sort=True)

        node_indices = nodes.index

        for id, data in self.components.pipes.iterrows():

            if not data['from_node'] in node_indices:
                raise ValueError(f"Node {data['from_node']} not defined.")

            if not data['to_node'] in node_indices:
                raise ValueError(f"Node {data['to_node']} not defined.")

            assert data['from_node'] != data['to_node'], \
                f"Pipe {id} connects {data['from_node']} to itself"

        if not self.components.pipes.empty:

            duplicate_pipes = [
                name for name, group in self.components.pipes.groupby(['from_node', 'to_node'])
                if len(group) > 1
            ]

            assert not duplicate_pipes, (
                f"There is more than one pipe that connects "
                f"{[pipe[0] + ' to ' + pipe[1] for pipe in duplicate_pipes]}")

        return True

    def _list_nested_dict_values(self, d):
        r"""
        Unwraps a nested dict and returns a list of all
        values in the branches of the dictionary.

        Parameters
        ----------
        d : dict
            Nested dictionary

        Returns
        -------
        leaves : list
            List of all values
        """
        leaves = []
        for _, v in d.items():
            if isinstance(v, dict):
                leaves.extend(self._list_nested_dict_values(v))
            else:
                leaves.append(v)
        return leaves

    @staticmethod
    def _are_indices_equal(indices):
        r"""
        Compares a list of pd.Index's and asserts
        that they are equal.

        Parameters
        ----------
        indices : list
            List containing pd.Index

        Returns
        -------
        True
        """
        if len(indices) == 1:
            print("Only one sequence given. Need more than one time-index to compare.")
            return True

        for index in indices[1:]:
            assert indices[0].equals(index)

        return True

    def set_timeindex(self):
        r"""
        Takes all sequences and checks if their timeindex is identical.
        If that is the case, it sets the timeindex attribute of the
        class.
        If there are no sequences given, the timeindex will keep the default value.
        """
        sequence_dfs = self._list_nested_dict_values(self.sequences)

        if sequence_dfs:

            indices = [df.index for df in sequence_dfs]

            self._are_indices_equal(indices)

            self.timeindex = indices[0]

        else:
            print("No sequences found to create timeindex from")

    def reproject(self, crs):
        pass

    def optimize_operation(self):
        self.results.operation = optimize_operation(self)

    def optimize_investment(self, invest_options, **kwargs):

        oemof_opti_model = setup_optimise_investment(
            self, invest_options, **kwargs
        )

        self.results.optimization = solve_optimisation_investment(
            oemof_opti_model
        )

    def simulate(self, *args, **kwargs):
        self.results.simulation = simulate(self, *args, **kwargs)



# -*- coding: utf-8

"""
This module is designed to hold optimization model implementations. The
implementation makes use of oemof-solph.

This file is part of project dhnx (). It's copyrighted
by the contributors recorded in the version control history of the file,
available from its original location:

SPDX-License-Identifier: MIT
"""


import oemof.solph as solph
import dhnx.optimization_oemof_heatpipe as oh


def add_buses(it, labels, nodes, busd):
    """
    This function initialises the oemof.solph.Bus classes of the energysystem based on the given
    tabular information. Additionally, a sink or a source can be added to every bus and costs for
    this source (shortage) or sink (excess) can be defined.

    The table must be given as follows:

    .. _bus_table:

    .. table:: Example for table of *Buses*

        +---------+--------+--------+----------+----------------+--------------+
        | label_2 | active | excess | shortage | shortage costs | excess costs |
        +=========+========+========+==========+================+==============+
        | heat    | 1      | 0      | 0        | 99999          | 99999        |
        +---------+--------+--------+----------+----------------+--------------+

    Parameters
    ----------
    it : pd.DateFrame
        Table of attributes for Buses for the producers and consumers.
    labels : dict
        Dictonary containing tag1 and tag4 of label-tuple.
    nodes : list
        All oemof.solph components are added to the list.
    busd : dict
        All buses are added to this dictionary.

    Returns
    -------
    list, dict : Updated list of nodes and dict of Buses.
    """

    for _, b in it.iterrows():

        labels['l_3'] = 'bus'
        labels['l_2'] = b['label_2']
        l_bus = oh.Label(labels['l_1'], labels['l_2'], labels['l_3'], labels['l_4'])

        # check if bus already exists (due to infrastructure)
        if l_bus in busd:
            print('Bus bereits vorhanden:', l_bus)

        else:
            bus = solph.Bus(label=l_bus)
            nodes.append(bus)

            busd[l_bus] = bus

            if b['excess']:
                labels['l_3'] = 'excess'
                nodes.append(
                    solph.Sink(label=oh.Label(
                        labels['l_1'], labels['l_2'], labels['l_3'], labels['l_4']),
                        inputs={busd[l_bus]: solph.Flow(variable_costs=b['excess costs'])}))

            if b['shortage']:
                labels['l_3'] = 'shortage'
                nodes.append(
                    solph.Source(label=oh.Label(
                        labels['l_1'], labels['l_2'], labels['l_3'], labels['l_4']),
                        outputs={busd[l_bus]: solph.Flow(variable_costs=b['shortage costs'])}))

    return nodes, busd


def add_sources(on, it, c, labels, nodes, busd):
    """
    The oemof.solph.Source components for the producers and consumers are initialised based on the
    given tabular information of the investment_options of the OemofInvestOptimizationModel.
    Time-series can also be used as attribute values for the outflow of the Source.
    Therefore, a table with the filename 'source_timeseries' must be given.

    Parameters
    ----------
    on : OemofInvestOptimizationModel
    it : DataFrame
        Table of attributes for Sources for the producers and consumers.
    c : Series
        Attributes of specific producer or consumer from the ThermalNetwork.
    labels : dict
        Dictonary containing specifications for label-tuple.
    nodes : list
        All oemof.solph components are added to the list.
    busd : dict
        All oemof.solph.Bus objects are given by this dictionary.

    Returns
    -------
    list : Updated list of nodes.
    """

    # check if timeseries are given
    ts_status = False   # status if timeseries for sources is present
    if 'source_' + 'timeseries' in on.invest_options[labels['l_1']].keys():
        ts = on.invest_options[labels['l_1']]['source_' + 'timeseries']
        ts_status = True

    # generel further flow attributes: check what flow attributes are given
    # by what comes after 'label_2'
    flow_attr = list(it.columns)
    idx = flow_attr.index('label_2')
    flow_attr = flow_attr[idx + 1:]

    for _, cs in it.iterrows():
        labels['l_3'] = 'source'
        labels['l_2'] = cs['label_2']

        outflow_args = {}

        # general additional flow attributes
        for fa in flow_attr:
            outflow_args[fa] = cs[fa]

        # specific flow attributes
        # e.g. check for heat (label 2)
        # e.g. check for source (label 3)
        spec_attr = [x for x in list(on.thermal_network.components[labels['l_1']].columns)
                     if x.split('.')[-1] in on.oemof_flow_attr
                     if x.split('.')[0] == labels['l_2']
                     if x.split('.')[1] == labels['l_3']]

        for sa in spec_attr:
            if sa.split('.')[-1] in outflow_args.keys():
                print(
                    'General attribute <{}> of Label 2 <{}> and Label 3 <{}> (value: {}) will '
                    'be replaced by specific data. New value for <{}>: {}'.format(
                        sa.split('.')[-1], labels['l_2'], labels['l_3'],
                        outflow_args[sa.split('.')[-1]], labels['l_4'], c[sa]))
            outflow_args[sa.split('.')[-1]] = c[sa]

        # add timeseries data if present
        if ts_status:
            ts_key = labels['l_4'].split('-')[1] + '_' + labels['l_2']
            for col in ts.columns.values:
                if col.split('.')[0] == ts_key:
                    outflow_args[col.split('.')[1]] = ts[col].values

        nodes.append(
            solph.Source(
                label=oh.Label(
                    labels['l_1'], labels['l_2'], labels['l_3'], labels['l_4']),
                outputs={
                    busd[(labels['l_1'], cs['label_2'], 'bus',
                          labels['l_4'])]: solph.Flow(**outflow_args)}))

    return nodes


def add_demand(it, labels, series, nodes, busd):
    """
    Initialisation of oemof.solph.Sinks which represent demands.

    Parameters
    ----------
    it : pd.DataFrame
        Table of attributes for Sources for the producers and consumers.
    labels : dict
        Dictonary containing specifications for label-tuple.
    series : dict
        Contain the heat demand time-series of all consumers.
    nodes : list
        All oemof.solph components are added to the list.
    busd : dict
        All oemof.solph.Bus objects are given by this dictionary.

    Returns
    -------
    list : Updated list of nodes.
    """

    for _, de in it.iterrows():
        labels['l_3'] = 'demand'
        labels['l_2'] = de['label_2']
        # set static inflow values
        inflow_args = {'nominal_value': de['nominal_value'],
                       'fix': series['heat_flow'][
                           labels['l_4'].split('-')[1]].values}

        # create
        nodes.append(
            solph.Sink(label=oh.Label(
                labels['l_1'], labels['l_2'], labels['l_3'], labels['l_4']),
                inputs={busd[(labels['l_1'], labels['l_2'], 'bus',
                              labels['l_4'])]: solph.Flow(**inflow_args)}))

    return nodes


def add_transformer(it, labels, nodes, busd):
    """Adds oemof.solph.Transformer objects to the list of components.

    If attribute `invest` is *True*, an `Investment` attribute is added to the outflow of the
    Transformer.

    Parameters
    ----------
    it : pd.DataFrame
        Table of transformer attributes of the producers / consumers.
    labels : dict
        Dictonary containing specifications for label-tuple.
    nodes : list
        All oemof.solph components are added to the list.
    busd : dict
        All oemof.solph.Bus objects are given by this dictionary.

    Returns
    -------
    list : Updated list of nodes.
    """

    for _, t in it.iterrows():
        labels['l_2'] = None
        labels['l_3'] = t['label_3']

        # Transformer with 1 Input and 1 Output
        if t['type'] == "1-in_1-out":

            b_in_1 = busd[(labels['l_1'], t['in_1'], 'bus', labels['l_4'])]
            b_out_1 = busd[(labels['l_1'], t['out_1'], 'bus',
                            labels['l_4'])]

            if t['invest']:

                if t['eff_out_1'] == 'series':
                    print('noch nicht angepasst!')

                epc_t = t['capex']

                # create
                nodes.append(
                    solph.Transformer(
                        label=oh.Label(
                            labels['l_1'], labels['l_2'], labels['l_3'], labels['l_4']),
                        inputs={b_in_1: solph.Flow()},
                        outputs={b_out_1: solph.Flow(
                            variable_costs=t['variable_costs'],
                            summed_max=t['in_1_sum_max'],
                            investment=solph.Investment(
                                ep_costs=epc_t + t['service'],
                                maximum=t['max_invest'],
                                minimum=t['min_invest']))},
                        conversion_factors={
                            b_out_1: t['eff_out_1']}))

            else:
                # create
                if t['eff_out_1'] == 'series':
                    print('noch nicht angepasst!')
                    # for col in nd['timeseries'].columns.values:
                    #     if col.split('.')[0] == t['label']:
                    #         t[col.split('.')[1]] = nd['timeseries'][
                    #             col]

                nodes.append(
                    solph.Transformer(
                        label=oh.Label(labels['l_1'], labels['l_2'], labels['l_3'],
                                       labels['l_4']),
                        inputs={b_in_1: solph.Flow()},
                        outputs={b_out_1: solph.Flow(
                            nominal_value=t['installed'],
                            summed_max=t['in_1_sum_max'],
                            variable_costs=t['variable_costs'])},
                        conversion_factors={b_out_1: t['eff_out_1']}))

    return nodes


def add_storage(it, labels, nodes, busd):
    """Adds oemof.solph.GenericStorage objects to the list of components.

    If attribute `invest` is *True*, the investment version of the Storage is created.

    Parameters
    ----------
    it : pd.DataFrame
        Table of storage attributes of the producers / consumers.
    labels : dict
        Dictonary containing specifications for label-tuple.
    nodes : list
        All oemof.solph components are added to the list.
    busd : dict
        All oemof.solph.Bus objects are given by this dictionary.

    Returns
    -------
    list : Updated list of nodes.
    """

    for _, s in it.iterrows():

        label_storage = oh.Label(labels['l_1'], s['bus'], s['label'], labels['l_4'])
        label_bus = busd[(labels['l_1'], s['bus'], 'bus', labels['l_4'])]

        if s['invest']:

            epc_s = s['capex']

            nodes.append(
                solph.components.GenericStorage(
                    label=label_storage,
                    inputs={label_bus: solph.Flow()},
                    outputs={label_bus: solph.Flow()},
                    loss_rate=s['capacity_loss'],
                    fixed_losses_relative=s['fixed_losses_relative'],
                    invest_relation_input_capacity=s[
                        'invest_relation_input_capacity'],
                    invest_relation_output_capacity=s[
                        'invest_relation_output_capacity'],
                    inflow_conversion_factor=s['inflow_conversion_factor'],
                    outflow_conversion_factor=s[
                        'outflow_conversion_factor'],
                    investment=solph.Investment(ep_costs=epc_s)))

        else:
            nodes.append(
                solph.components.GenericStorage(
                    label=label_storage,
                    inputs={label_bus: solph.Flow()},
                    outputs={label_bus: solph.Flow()},
                    loss_rate=s['capacity_loss'],
                    fixed_losses_relative=s['fixed_losses_relative'],
                    nominal_storage_capacity=s['capacity'],
                    inflow_conversion_factor=s['inflow_conversion_factor'],
                    outflow_conversion_factor=s[
                        'outflow_conversion_factor']))

    return nodes


def add_heatpipes(it, labels, gd, q, b_in, b_out, nodes):
    """
    Adds *HeatPipeline* objects with *Investment* attribute to the list of oemof.solph components.

    Parameters
    ----------
    it : pd.DataFrame
        Table of *Heatpipeline* attributes of the district heating grid
    labels : dict
        Dictonary containing specifications for label-tuple
    gd : dict
        Settings of the investment optimisation of the ThermalNetwork
    q : pd.Series
        Specific *Pipe* of ThermalNetwork
    b_in : oemof.solph.Bus
        Bus of Inflow
    b_out : oemof.solph.Bus
        Bus of Outflow
    nodes : list
        All oemof.solph components are added to the list

    Returns
    -------
    list : Updated list of nodes.
    """

    for _, t in it.iterrows():

        # definition of tag3 of label -> type of pipe
        labels['l_3'] = t['label_3']

        epc_p = t['capex_pipes'] * q['length']
        epc_fix = t['fix_costs'] * q['length']

        # Heatpipe with binary variable
        nc = bool(t['nonconvex'])

        # bidirectional heatpipelines yes or no
        flow_bi_args = {
            'bidirectional': True, 'min': -1}\
            if gd['bidirectional_pipes'] else {}

        nodes.append(oh.HeatPipeline(
            label=oh.Label(labels['l_1'], labels['l_2'],
                           labels['l_3'], labels['l_4']),
            inputs={b_in: solph.Flow(**flow_bi_args)},
            outputs={b_out: solph.Flow(
                nominal_value=None,
                **flow_bi_args,
                investment=solph.Investment(
                    ep_costs=epc_p, maximum=t['cap_max'],
                    minimum=t['cap_min'], nonconvex=nc, offset=epc_fix,
                ))},
            heat_loss_factor=t['l_factor'] * q['length'],
            heat_loss_factor_fix=t['l_factor_fix'] * q['length'],
        ))

    return nodes


def add_heatpipes_exist(pipes, labels, gd, q, b_in, b_out, nodes):
    """
    Adds *HeatPipeline* objects with fix capacity for existing pipes
    to the list of oemof.solph components.

    Parameters
    ----------
    pipes
    labels : dict
        Dictonary containing specifications for label-tuple.
    gd : dict
        Settings of the investment optimisation of the ThermalNetwork
    q : pd.Series
        Specific *Pipe* of ThermalNetwork
    b_in : oemof.solph.Bus
        Bus of Inflow
    b_out : oemof.solph.Bus
        Bus of Outflow
    nodes : list
        All oemof.solph components are added to the list

    Returns
    -------
    list : Updated list of nodes.
    """

    # get index of existing pipe label of pipe data
    ind_pipe = pipes[pipes['label_3'] == q['hp_type']].index[0]
    t = pipes.loc[ind_pipe]
    # get label of pipe
    labels['l_3'] = t['label_3']

    hlf = t['l_factor'] * q['length']
    hlff = t['l_factor_fix'] * q['length']

    flow_bi_args = {
        'bidirectional': True, 'min': -1} \
        if gd['bidirectional_pipes'] else {}

    outflow_args = {'nonconvex': solph.NonConvex()} if t['nonconvex'] else {}

    nodes.append(oh.HeatPipeline(
        label=oh.Label(labels['l_1'], labels['l_2'],
                       labels['l_3'], labels['l_4']),
        inputs={b_in: solph.Flow(**flow_bi_args)},
        outputs={b_out: solph.Flow(
            nominal_value=q['capacity'],
            **flow_bi_args,
            **outflow_args,
        )},
        heat_loss_factor=hlf,
        heat_loss_factor_fix=hlff,
    ))

    return nodes



# -*- coding: utf-8

"""
This module is designed to hold functions for visualization.

This file is part of project dhnx (). It's copyrighted
by the contributors recorded in the version control history of the file,
available from its original location:

SPDX-License-Identifier: MIT
"""

from collections import namedtuple
import logging

import folium as fol
from folium.features import DivIcon
import matplotlib.pyplot as plt
import matplotlib.collections as collections
import numpy as np
import pandas as pd

logger = logging.getLogger()
logger.setLevel(logging.INFO)

cartopy_installed = True

try:
    from cartopy.io.img_tiles import Stamen
    from cartopy import crs as ccrs

except ImportError:
    logging.info("Cartopy is not installed. Background maps will not be drawn.")
    cartopy_installed = False


class InteractiveMap():
    r"""
    An interactive map of a network.ThermalNetwork.
    """
    def __init__(self, thermal_network):
        self.node_data = self.collect_node_data(thermal_network)
        self.edge_data = thermal_network.components.pipes
        self.edge_data['value'] = 1
        self.node_id = self.node_data.index
        self.lat = self.node_data['lat']
        self.lon = self.node_data['lon']
        self.component_type = self.node_data['component_type']
        self._add_colors()

    @staticmethod
    def collect_node_data(thermal_network):
        node_data = {
            list_name: thermal_network.components[list_name].copy() for list_name in [
                'consumers',
                'producers',
                'forks'
            ]
        }

        for k, v in node_data.items():
            v.index = [k + '-' + str(id) for id in v.index]

        return pd.concat(node_data.values())

    def _add_colors(self):
        color = {'producer': '#ff0000',
                 'consumer': '#00ff00',
                 'split': '#000000'}

        self.node_data = (
            self.node_data
            .assign(node_color=self.node_data['component_type'])
            .replace({'node_color': color}))

        return self.node_data['node_color']

    @staticmethod
    def _get_bearing(p1, p2):
        '''
        Returns compass bearing from p1 to p2

        Parameters
        p1 : namedtuple with lat lon
        p2 : namedtuple with lat lon

        Return
        compass bearing of type float
        '''
        y = p2[0] - p1[0]
        x = p2[1] - p1[1]

        bearing = np.arctan2(x, y) / np.pi * 180

        # adjusting for compass bearing
        if bearing < 0:
            return bearing + 360

        return bearing

    def _get_arrows(self, locations, color='black', size=8, n_arrows=3):
        '''
        Get a list of correctly placed and rotated
        arrows/markers to be plotted

        Parameters
        locations : list of lists of lat lons that represent the
                    start and end of the line.
                    eg [[41.1132, -96.1993],[41.3810, -95.8021]]
        color : default is 'black'
        size : default is 8
        n_arrows : number of arrows to create.  default is 3

        Return
        list of arrows/markers
        '''

        Point = namedtuple('Point', field_names=['lat', 'lon'])

        # creating point from our Point named tuple
        p1 = Point(locations[0][0], locations[0][1])
        p2 = Point(locations[1][0], locations[1][1])

        # getting the rotation needed for our marker.
        # Subtracting 90 to account for the marker's orientation
        # of due East(get_bearing returns North)
        rotation = self._get_bearing(p1, p2) - 90

        # get an evenly space list of lats and lons for our arrows
        # note that I'm discarding the first and last for aesthetics
        # as I'm using markers to denote the start and end
        arrow_lats = np.linspace(p1.lat, p2.lat, n_arrows + 2)[1:n_arrows + 1]
        arrow_lons = np.linspace(p1.lon, p2.lon, n_arrows + 2)[1:n_arrows + 1]

        arrows = []

        #creating each "arrow" and appending them to our arrows list
        for points in zip(arrow_lats, arrow_lons):
            arrows.append(
                fol.RegularPolygonMarker(
                    location=points,
                    color=color, number_of_sides=3,
                    radius=size, rotation=rotation, fill=True))

        return arrows

    def draw(self):
        # create map
        m = fol.Map(location=[self.lat.mean(), self.lon.mean()],
                    zoom_start=14)

        for i in range(0, len(self.node_data)):
            # draw nodes
            fol.CircleMarker([self.lat[i], self.lon[i]],
                             # popup=data['node_id'][i],
                             color=self.node_data['node_color'][i],
                             fill_color=self.node_data['node_color'][i],
                             radius=20).add_to(m)

            # draw node ids
            fol.Marker(
                [self.lat[i], self.lon[i]],
                icon=DivIcon(
                    icon_size=(-35, 75),
                    icon_anchor=(0, 0),
                    html='<div style="font-size: 16pt">%s</div>'
                    % self.node_data.index[i]
                )
            ).add_to(m)

        for i in range(0, len(self.edge_data)):
            # linewidth settings
            lw_avg = self.edge_data['value'].mean()
            lw = self.edge_data['value'][i] / lw_avg

            fol.PolyLine(locations=[[self.lat[self.edge_data['from_node'][i]],
                                     self.lon[self.edge_data['from_node'][i]]],
                                    [self.lat[self.edge_data['to_node'][i]],
                                     self.lon[self.edge_data['to_node'][i]]]],
                         color='orange',
                         weight=lw * 3).add_to(m)

            arrows = self._get_arrows(
                locations=[[self.lat[self.edge_data['from_node'][i]],
                            self.lon[self.edge_data['from_node'][i]]],
                           [self.lat[self.edge_data['to_node'][i]],
                            self.lon[self.edge_data['to_node'][i]]]],
                color='orange', n_arrows=3)

            for arrow in arrows:
                arrow.add_to(m)

        return m


class StaticMap():
    r"""
    A static map of a network.ThermalNetwork.
    """
    def __init__(self, thermal_network, figsize=(5, 5), node_size=3,
                 edge_width=3, node_color='r', edge_color='g'):
        self.graph = thermal_network.to_nx_graph()
        self.figsize = figsize
        self.node_size = node_size
        self.edge_width = edge_width
        self.node_color = node_color
        self.edge_color = edge_color
        self.positions = {node_id: np.array([data['lon'], data['lat']])
                          for node_id, data in self.graph.nodes(data=True)}
        self.extent = self._get_extent()

    def _get_extent(self):
        lon = [pos[0] for pos in self.positions.values()]
        lat = [pos[1] for pos in self.positions.values()]
        extent = np.array([np.min(lon), np.max(lon), np.min(lat), np.max(lat)])
        delta = [extent[1] - extent[0], extent[3] - extent[2]]
        extent = extent.astype(float)
        extent += 0.1 * np.array([-delta[0], delta[0], -delta[1], delta[1]])
        return extent

    def draw(self, bgcolor='w', no_axis=False, background_map=False,
             use_geom=False, edge_color='b', edge_linewidth=2,
             edge_alpha=1, node_size=40, node_color='r', node_alpha=1,
             edgecolor='r', node_zorder=1):
        """
        This function has been adapted from osmnx plots.plot_graph() function.
        """
        if background_map:
            if not cartopy_installed:
                logging.warning('To draw background map, cartopy must be installed.')
                background_map = False

        if background_map:
            imagery = Stamen(style='toner-lite')
            zoom_level = 15
            fig, ax = plt.subplots(
                figsize=self.figsize,
                subplot_kw={'projection': imagery.crs}
            )
            ax.set_extent(self.extent, crs=ccrs.Geodetic())
            ax.add_image(imagery, zoom_level, alpha=1, interpolation='bilinear')

        else:
            fig, ax = plt.subplots(figsize=self.figsize, facecolor=bgcolor)

        lines = []
        for u, v, data in self.graph.edges(data=True):
            if 'geometry' in data and use_geom:
                # if it has a geometry attribute (a list of line segments), add them
                # to the list of lines to plot
                xs, ys = data['geometry'].xy
                lines.append(list(zip(xs, ys)))
            else:
                # if it doesn't have a geometry attribute, the edge is a straight
                # line from node to node
                x1 = self.graph.nodes[u]['lon']
                y1 = self.graph.nodes[u]['lat']
                x2 = self.graph.nodes[v]['lon']
                y2 = self.graph.nodes[v]['lat']
                line = [(x1, y1), (x2, y2)]
                lines.append(line)

        # add the lines to the axis as a linecollection
        lc = collections.LineCollection(lines,
                                        colors=edge_color,
                                        linewidths=edge_linewidth,
                                        alpha=edge_alpha,
                                        zorder=2)
        ax.add_collection(lc)

        node_Xs = [float(x) for _, x in self.graph.nodes(data='lon')]
        node_Ys = [float(y) for _, y in self.graph.nodes(data='lat')]

        ax.scatter(node_Xs,
                   node_Ys,
                   s=node_size,
                   c=node_color,
                   alpha=node_alpha,
                   edgecolor=edgecolor,
                   zorder=node_zorder)

        if no_axis:
            ax = plt.gca()
            ax.set_axis_off()

        return fig, ax



# -*- coding: utf-8

"""
This module is designed to hold implementations of simulation models. The
implementation uses oemof/tespy.

This file is part of project dhnx (). It's copyrighted
by the contributors recorded in the version control history of the file,
available from its original location:

SPDX-License-Identifier: MIT
"""
import re
import warnings
from functools import partial

import networkx as nx
import numpy as np
import pandas as pd

from .model import SimulationModel
from .helpers import Dict, sum_ignore_none
from .graph import write_edge_data_to_graph
from .input_output import save_results


idx = pd.IndexSlice


class SimulationModelNumpy(SimulationModel):
    r"""
    Implementation of a simulation model using numpy.
    """
    def __init__(
            self, thermal_network,
            rho=971.78, c=4190, mu=0.00035, eta_pump=1, tolerance=1e-10
    ):
        super().__init__(thermal_network)
        self.results = {}

        self.nx_graph = thermal_network.to_nx_graph()

        assert nx.algorithms.tree.is_tree(self.nx_graph),\
            "Currently, only tree networks can be modeled. " \
            "Looped networks are not implemented yet."

        self.inc_mat = nx.incidence_matrix(self.nx_graph, oriented=True).todense()

        self.input_data = Dict()

        self.rho = rho  # kg/m3

        self.c = c  # J/(kg*K)

        self.mu = mu  # kg/(m*s)

        self.temp_env = thermal_network.sequences.environment.temp_env.iloc[:, 0]

        self.eta_pump = eta_pump

        self.tolerance = tolerance

        if self._concat_scalars('height') is not None:
            warnings.warn(
                "Pressure differences due to height differences are not implemented yet."
            )

    def prepare(self):

        self.prepare_hydraulic_eqn()

        self.prepare_thermal_eqn()

    def solve(self):

        self.solve_hydraulic_eqn()

        self.solve_thermal_eqn()

    def get_results(self):

        return self.results

    def prepare_hydraulic_eqn(self):
        r"""
        Prepares the input data for the hydraulic problem.
        """
        self.input_data.mass_flow = pd.DataFrame(
            0,
            columns=self.nx_graph.nodes(),
            index=self.thermal_network.timeindex
        )

        input_data = self._concat_sequences('mass_flow')

        self.input_data.mass_flow.loc[:, input_data.columns] = input_data

        self.input_data.mass_flow = self._set_producers_mass_flow(self.input_data.mass_flow)

    def prepare_thermal_eqn(self):
        r"""
        Prepares the input data for the thermal problem.
        """

        self.input_data.temp_inlet = pd.DataFrame(
            0,
            columns=self.nx_graph.nodes(),
            index=self.thermal_network.timeindex
        )

        input_data = self._concat_sequences('temp_inlet')

        self.input_data.temp_inlet.loc[:, input_data.columns] = input_data

    def solve_hydraulic_eqn(self):
        r"""
        Solves the hydraulic problem.
        """
        self.results['pipes-mass_flow'] = self._calculate_pipes_mass_flow()

        reynolds = self._calculate_reynolds()

        lamb = self._calculate_lambda(reynolds)

        pipes_dist_pressure_losses = self._calculate_pipes_distributed_pressure_losses(lamb)

        pipes_loc_pressure_losses = self._calculate_pipes_localized_pressure_losses()

        pipes_total_pressure_losses = sum_ignore_none(
            pipes_dist_pressure_losses, pipes_loc_pressure_losses
        )

        global_pressure_losses = self._calculate_global_pressure_losses(pipes_total_pressure_losses)

        pump_power = self._calculate_pump_power(global_pressure_losses)

        self.results['pipes-dist_pressure_losses'] = pipes_dist_pressure_losses

        self.results['pipes_loc_pressure_losses'] = pipes_loc_pressure_losses

        self.results['global-pressure_losses'] = global_pressure_losses

        self.results['producers-pump_power'] = pump_power

    def solve_thermal_eqn(self):
        r"""
        Solves the thermal problem.
        """
        exponent_constant = self._calculate_exponent_constant()

        temp_inlet = self._calc_temps(exponent_constant, self.input_data.temp_inlet, direction=1)

        temp_return_known = self._set_temp_return_input(temp_inlet)

        temp_return = self._calc_temps(exponent_constant, temp_return_known, direction=-1)

        pipes_heat_losses = self._calculate_pipes_heat_losses(temp_inlet) \
            + self._calculate_pipes_heat_losses(temp_return)

        global_heat_losses = pipes_heat_losses.sum(axis=1)

        global_heat_losses.name = 'global_heat_losses'

        self.results['nodes-temp_inlet'] = temp_inlet

        self.results['nodes-temp_return'] = temp_return

        self.results['pipes-heat_losses'] = pipes_heat_losses

        self.results['global-heat_losses'] = global_heat_losses

    def _concat_scalars(self, name):
        r"""
        Concatenates scalars of all components with a given variable name

        Parameters
        ----------
        name : str
            Name of the variable

        Returns
        -------
        concat_sequences : pd.DataFrame
            DataFrame containing the sequences
        """
        select_scalars = [
            scalar[name].copy().rename(index=lambda x, prefix=component: prefix + '-' + str(x))
            for component, scalar in self.thermal_network.components.items()
            if name in scalar
        ]

        if select_scalars:
            select_scalars = pd.concat(select_scalars, 0)

        else:
            select_scalars = None

        return select_scalars

    def _concat_sequences(self, name):
        r"""
        Concatenates sequences of all components with a given variable name

        Parameters
        ----------
        name : str
            Name of the variable

        Returns
        -------
        concat_sequences : pd.DataFrame
            DataFrame containing the sequences
        """
        select_sequences = [
            d[name].copy().rename(columns=lambda x, prefix=component: prefix + '-' + x)
            for component, d in self.thermal_network.sequences.items()
            if name in d
        ]

        concat_sequences = pd.concat(select_sequences, 1)

        return concat_sequences

    @staticmethod
    def _set_producers_mass_flow(m):
        r"""
        Sets the mass flow of the producer.

        Parameters
        ----------
        m : pd.DataFrame
            DataFrame with all know consumer mass flows.

        Returns
        -------
        m : pd.DataFrame
            DataFrame with all know mass flow of
            consumers and producer.
        """
        producers = [name for name in m.columns if re.search('producers', name)]

        assert len(producers) == 1, "Currently, only one producer allowed."

        m.loc[:, producers] = - m.loc[:, ~m.columns.isin(producers)].sum(1)

        return m

    def _calculate_pipes_mass_flow(self):
        r"""
        Determines the mass flow in all pipes using numpy's
        least squares function.

        Returns
        -------
        pipes_mass_flow : pd.DataFrame
            Mass flow in the pipes [kg/s]
        """

        pipes_mass_flow = {}

        for t in self.thermal_network.timeindex:

            # The order of columns in self.input_data.mass_flow fit with those of self.inc_mat
            # because the columns have been generated from the graph's nodes in
            # prepare_hydraulic_eqn()

            x, residuals, _, _ = np.linalg.lstsq(
                self.inc_mat,
                self.input_data.mass_flow.loc[t, :],
                rcond=None
            )

            assert residuals < self.tolerance,\
                f"Residuals {residuals} are larger than tolerance {self.tolerance}!"

            pipes_mass_flow.update({t: x})

        pipes_mass_flow = pd.DataFrame.from_dict(
            pipes_mass_flow,
            orient='index',
            columns=self.nx_graph.edges()
        )

        pipes_mass_flow.columns.names = ('from_node', 'to_node')

        return pipes_mass_flow

    def _calculate_reynolds(self):
        r"""
        Calculates the Reynolds number.

        .. math::

            Re = \frac{4\dot{m}}{\pi\mu D}

        Returns
        -------
        re : pd.DataFrame
            Reynolds number for every time step and pipe [-]
        """
        pipes_mass_flow = self.results['pipes-mass_flow']

        diameter = \
            self.thermal_network.components.pipes[['from_node', 'to_node', 'diameter']]

        diameter = 1e-3 * diameter.set_index(['from_node', 'to_node'])['diameter']

        reynolds = 4 * pipes_mass_flow.divide(diameter, axis='columns') \
            / (np.pi * self.mu)

        return reynolds

    def _calculate_lambda(self, reynolds):
        r"""
        Calculates the darcy friction factor.

        .. math::

            \lambda = 0.07 \cdot Re ^ {-0.13} \cdot D ^ {-0.14}

        Parameters
        ----------
        re : pd.DataFrame
            Reynolds number for every time step and pipe [-]

        Returns
        -------
        lamb : pd.DataFrame
            Darcy friction factor for every time step and pipe [-]
        """

        factor_diameter = self.thermal_network.components.pipes[
            ['from_node', 'to_node', 'diameter']
        ]

        factor_diameter = (
            1e-3 * factor_diameter.set_index(['from_node', 'to_node'])['diameter']
        ) ** -0.14

        lamb = 0.07 * reynolds ** -0.13

        lamb = lamb.multiply(factor_diameter, axis='columns')

        return lamb

    def _calculate_pipes_distributed_pressure_losses(self, lamb):
        r"""
        Calculates the pressure losses in the pipes.

        Equal-sized inlet and return pipes are assumed which leads to
        equal mass flows and pressure losses for both. This introduces
        the initial factor of 2 in the equation.

        .. math::

            \delta p = 2 \cdot \lambda \frac{8L}{\rho \pi^2 D^5}\dot{m}^2.

        Parameters
        ----------
        lamb : pd.DataFrame
            Darcy friction factor for every time step and pipe [-]

        Returns
        -------
        pipes_pressure_losses : pd.DataFrame
            DataFrame with distributed pressure losses for inlet and return for every
            time step and pipe [Pa]
        """
        pipes_mass_flow = self.results['pipes-mass_flow'].copy()

        pipes_mass_flow_2 = pipes_mass_flow ** 2

        constant = 8 * lamb / (self.rho * np.pi**2)

        length = self.thermal_network.components.pipes[['from_node', 'to_node', 'length']]

        diameter = \
            self.thermal_network.components.pipes[['from_node', 'to_node', 'diameter']]

        length = length.set_index(['from_node', 'to_node'])['length']

        diameter = diameter.set_index(['from_node', 'to_node'])['diameter']

        diameter_5 = (1e-3 * diameter) ** 5

        pipes_pressure_losses = constant * pipes_mass_flow_2\
            .multiply(length, axis='columns')\
            .divide(diameter_5, axis='columns')

        # We multiply by the factor of two to represent the pressure losses along inlet
        # and return flow.

        pipes_pressure_losses *= 2

        return pipes_pressure_losses

    def _calculate_pipes_localized_pressure_losses(self):
        r"""
        Calculates localized pressure losses at the nodes.

        .. math::

            \Delta p_{loc} = \frac{8\zeta\dot{m}^2}{\rho \pi^2 D^4}

        Returns
        -------
        nodes_pressure_losses : pd.DataFrame
            Localized pressure losses at the nodes [Pa]
        """
        def _assign_zeta_to_pipes(flow_type):

            def func(row, level):

                res = zeta.reindex(row.index.get_level_values(level).values)

                res.index = row.index

                return res

            zeta = self._concat_scalars('zeta_' + flow_type)

            if zeta is None:
                print(f"No values for zeta_{flow_type} found. Skipping.")

                return None

            flow_direction = np.sign(self.results['pipes-mass_flow'])

            zeta_pipes = pd.DataFrame(
                columns=flow_direction.columns,
                index=flow_direction.index
            )

            zeta_pipes_forward = zeta_pipes.copy()

            zeta_pipes_backward = zeta_pipes.copy()

            func_forward = partial(func, level=0)

            zeta_pipes_forward = zeta_pipes_forward.apply(func_forward, axis=1)

            func_backward = partial(func, level=1)

            zeta_pipes_backward = zeta_pipes_backward.apply(func_backward, axis=1)

            if flow_type == 'inlet':
                zeta_pipes[flow_direction > 0] = zeta_pipes_forward

                zeta_pipes[flow_direction < 0] = zeta_pipes_backward

            elif flow_type == 'return':
                zeta_pipes[flow_direction > 0] = zeta_pipes_backward

                zeta_pipes[flow_direction < 0] = zeta_pipes_forward

            else:
                raise ValueError("Flow type has to be either inlet or return.")

            return zeta_pipes

        def _calc_loc_pressure_loss_for_flow_type(flow_type):
            zeta_pipes = _assign_zeta_to_pipes(flow_type)

            if zeta_pipes is None:
                return None

            pipes_localized_pressure_losses = {}

            for t in self.thermal_network.timeindex:
                pipes_mass_flow = self.results['pipes-mass_flow'].loc[t, :]

                mass_flow_2 = pipes_mass_flow ** 2

                mass_flow_2_over_diameter_4 = mass_flow_2.divide(diameter_4)

                mass_flow_2_over_diameter_4.name = 'mass_flow_2_over_diameter_4'

                x = constant * zeta_pipes.loc[t, :].multiply(mass_flow_2_over_diameter_4)

                pipes_localized_pressure_losses.update({t: x})

            pipes_localized_pressure_losses = pd.DataFrame.from_dict(
                pipes_localized_pressure_losses,
                orient='index'
            )

            pipes_localized_pressure_losses = pipes_localized_pressure_losses.fillna(0)

            return pipes_localized_pressure_losses

        constant = 8 / (self.rho * np.pi ** 2)

        diameter_4 = self.thermal_network.components.pipes[
            ['from_node', 'to_node', 'diameter']
        ]

        diameter_4 = diameter_4.set_index(['from_node', 'to_node'])['diameter']

        diameter_4 = (1e-3 * diameter_4) ** 4

        pipes_localized_pressure_losses_inlet = _calc_loc_pressure_loss_for_flow_type('inlet')

        pipes_localized_pressure_losses_return = _calc_loc_pressure_loss_for_flow_type('return')

        pipes_localized_pressure_losses = sum_ignore_none(
            pipes_localized_pressure_losses_inlet, pipes_localized_pressure_losses_return
        )

        return pipes_localized_pressure_losses

    def _calculate_global_pressure_losses(self, pipes_pressure_losses):
        r"""
        Calculates global pressure losses.

        Finds the path with the maximal pressure loss among from the set of
        paths from the producer to all consumers.

        Parameters
        ----------
        pipes_pressure_losses : pd.DataFrame
            Total pressure losses for every time step and pipe [Pa]

        Returns
        -------
         global_pressure_losses : pd.DataFrame
            Global pressure losses [Pa]
        """
        def calculate_path_weights(source_nodes, sink_nodes, graph, weight):

            path_weights = {}

            for sink in sink_nodes:
                for source in source_nodes:

                    path_weights[sink] = nx.dijkstra_path_length(
                        graph,
                        source=source,
                        target=sink,
                        weight=weight
                    )

            return path_weights

        def _calculate_paths_pressure_losses():

            sink_nodes = [node for node, data in self.nx_graph.nodes(data=True) if
                          data['component_type'] == 'Consumer']

            source_nodes = [node for node, data in self.nx_graph.nodes(data=True) if
                            data['component_type'] == 'Producer']

            paths_pressure_losses = {}

            for t in self.thermal_network.timeindex:

                graph = write_edge_data_to_graph(
                    pipes_pressure_losses.loc[t, :],
                    self.nx_graph,
                    var_name='pipes_pressure_losses'
                )

                path_weights = calculate_path_weights(
                    source_nodes,
                    sink_nodes,
                    graph,
                    'pipes_pressure_losses'
                )

                paths_pressure_losses[t] = path_weights

            paths_pressure_losses = \
                pd.DataFrame.from_dict(paths_pressure_losses, orient='index')

            return paths_pressure_losses

        paths_pressure_losses = _calculate_paths_pressure_losses()

        # Here, we take the path with the maximum pressure losses and assume that the other
        # consumer's valves are adjusted so that in sum, the pressure losses along all paths are
        # equal.

        global_pressure_losses = paths_pressure_losses.max(axis=1)

        return global_pressure_losses

    def _calculate_pump_power(self, global_pressure_losses):
        r"""
        Calculates the pump power.

        .. math::

            P_{el. pump} = \frac{1}{\eta_{el}\eta_{hyd}}\frac{\Delta p }{\rho} \dot{m}

        Parameters
        ----------
        global_pressure_losses : pd.DataFrame
            Global pressure losses [Pa]

        Returns
        -------
         pump_power : pd.Series
            Pump power [W]
        """
        producers = [
            node for node, data in self.nx_graph.nodes(data=True)
            if data['component_type'] == 'Producer'
        ]

        mass_flow_producers = \
            self.results['pipes-mass_flow'].loc[:, idx[producers, :]].sum(axis=1)

        pump_power = mass_flow_producers * global_pressure_losses / (self.eta_pump * self.rho)

        return pump_power

    def _calculate_exponent_constant(self):
        r"""
        Calculates the constant part of the exponent that determines the
        cooling of the medium in the pipes.

        .. math::

            exp_{const} = - \frac{U \pi D L }{c}

        Returns
        -------
        exponent_constant : np.matrix
            Constant part of the exponent [kg/s]
        """

        heat_transfer_coefficient = nx.adjacency_matrix(
            self.nx_graph, weight='heat_transfer_coefficient').todense()

        diameter = 1e-3 * nx.adjacency_matrix(self.nx_graph, weight='diameter').todense()

        length = nx.adjacency_matrix(self.nx_graph, weight='length').todense()

        exponent_constant = - np.pi \
            * np.multiply(heat_transfer_coefficient, np.multiply(diameter, length)) \
            / self.c

        return exponent_constant

    def _calc_temps(self, exponent_constant, known_temp, direction):
        r"""
        Calculate temperatures

        .. math::

            T_{out} = T_{env} + (T_{in} - T_{env}) \cdot exp\{exp_{const} \cdot exp_{var}\} =
            T_{out} = T_{env} + (T_{in} - T_{env}) \cdot exp\{-\frac{U \pi D L}{c \cdot \dot{m}}\}

        Parameters
        ----------
        exponent_constant : np.array
            Constant part of the exponent [kg/s]

        known_temp : pd.DataFrame
            Known temperatures at producers or consumers [°C]

        direction : +1 or -1
            For inlet and return flow [-]

        Returns
        -------
        temp_df : pd.DataFrame
            DataFrame containing temperatures for all nodes [°C]
        """
        # TODO: Rethink function layout and naming

        temps = {}

        for t in self.thermal_network.timeindex:

            # Divide exponent matrix by current pipes-mass_flows.
            data = self.results['pipes-mass_flow'].loc[t, :].copy()

            data = 1 / data

            graph_with_data = write_edge_data_to_graph(
                data, self.nx_graph, var_name='pipes-mass_flow'
            )

            inverse_mass_flows = nx.adjacency_matrix(
                graph_with_data, weight='pipes-mass_flow'
            ).todense()

            exponent = np.multiply(exponent_constant, inverse_mass_flows)

            matrix = np.exp(exponent)

            # Clear out elements where matrix was zero before exponentiation. This could be
            # replaced by properly passing `where` to np.multiply in the line above.
            matrix = np.multiply(matrix, nx.adjacency_matrix(self.nx_graph).todense())

            # Adapt matrix
            if direction == 1:
                matrix = matrix.T
                normalisation = np.array(nx.adjacency_matrix(self.nx_graph).sum(0)).flatten()

                normalisation = \
                    np.divide(np.array([1]), normalisation, where=normalisation != 0)

            elif direction == -1:
                normalisation = np.array(nx.adjacency_matrix(self.nx_graph).sum(1)).flatten()

                normalisation = \
                    np.divide(np.array([1]), normalisation, where=normalisation != 0)

            else:
                raise ValueError("Direction has to be either 1 or -1.")

            normalisation = np.diag(normalisation)

            matrix = np.dot(normalisation, matrix)

            matrix = np.identity(len(matrix)) - matrix

            vector = np.array(known_temp.loc[t])

            vector[vector != 0] -= self.temp_env.loc[t]

            x, _, _, _ = np.linalg.lstsq(
                matrix,
                vector,
                rcond=None
            )

            temps.update({t: x + self.temp_env.loc[t]})

        temp_df = pd.DataFrame.from_dict(
            temps,
            orient='index',
            columns=self.nx_graph.nodes()
        )

        return temp_df

    def _set_temp_return_input(self, temp_inlet):
        r"""
        Sets the temperature of the return pipes
        at the consumers.

        .. math::

            T_{cons,r} = T_{cons,i} - T_{cons,drop}

        Parameters
        ----------
        temp_inlet : pd.DataFrame
            Known inlet temperature [°C]

        Returns
        -------
        temp_return : pd.DataFrame
            Return temperature with the consumers values set [°C]
        """

        temp_return = pd.DataFrame(
            0,
            columns=self.nx_graph.nodes(),
            index=self.thermal_network.timeindex
        )

        temp_drop = self._concat_sequences('delta_temp_drop')

        temp_return.loc[:, temp_drop.columns] = temp_inlet.loc[:, temp_drop.columns] - temp_drop

        return temp_return

    def _calculate_pipes_heat_losses(self, temp_node):
        r"""
        Calculates the pipes' heat losses given the
        temperatures.

        .. math::

            \dot{Q}_{losses} = c \cdot \dot{m} \cdot \Delta T

        Parameters
        ----------
        temp_node : pd.DataFrame
            Temperatures at the nodes [°C]

        Returns
        -------
        pipes_heat_losses : pd.DataFrame
            Heat losses in the pipes [W]
        """

        pipes_heat_losses = {}

        for i, row in temp_node.iterrows():

            mass_flow = self.results['pipes-mass_flow'].loc[i, :].copy()

            temp_difference = np.abs(np.array(np.dot(row, self.inc_mat)).flatten())

            pipes_heat_losses[i] = self.c * mass_flow.multiply(temp_difference, axis=0)

        pipes_heat_losses = pd.DataFrame.from_dict(pipes_heat_losses, orient='index')

        return pipes_heat_losses


def simulate(thermal_network, results_dir=None):
    r"""
    Takes a thermal network and returns the result of
    the simulation.

    Parameters
    ----------
    thermal_network

    Returns
    -------
    results : dict
    """
    model = SimulationModelNumpy(thermal_network)

    model.prepare()

    model.solve()

    results = model.get_results()

    if results_dir is not None:
        save_results(results, results_dir)

    return results



# -*- coding: utf-8

"""
This module is designed to hold optimization model implementations. The
implementation makes use of oemof-solph.

This file is part of project dhnx (). It's copyrighted
by the contributors recorded in the version control history of the file,
available from its original location:

SPDX-License-Identifier: MIT
"""
# pylint: skip-file
import warnings
from collections import namedtuple

from pyomo.core.base.block import SimpleBlock
from pyomo.environ import (Set, NonNegativeReals, Var, Constraint)

from oemof.solph.network import Transformer
from oemof.solph.plumbing import sequence
from oemof.solph import Investment


class Label(namedtuple('solph_label', ['tag1', 'tag2', 'tag3', 'tag4'])):
    __slots__ = ()

    def __str__(self):
        """The string is used within solph as an ID, so it hast to be unique"""
        return '_'.join(map(str, self._asdict().values()))


class HeatPipeline(Transformer):
    r"""A HeatPipeline represent a Pipeline in a district heating system.
    This is done by a Transformer with a constant energy loss independent of
    actual power, but dependent on the nominal power and the length parameter.
    The HeatPipeline is a single-input-single-output transformer. Additionally,
    conversion factors for in- and output flow can be applied.

    Parameters
    ----------
    length : float
        Length of HeatPipeline.
    heat_loss_factor : float
        Heat loss per length unit as fraction of the nominal power. Can also be
        defined by a series.

    See also :py:class:`~oemof.solph.network.Transformer`.

    Note
    ----
    This component is experimental. Use it with care.


    The following sets, variables, constraints and objective parts are created
     * :py:class:`~oemof.solph.custom.HeatPipelineBlock` (if no
       Investment object present)
     * :py:class:`~oemof.solph.custom.HeatPipelineInvestBlock` (if
       Investment object present)

    Examples
    --------
    example

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.heat_loss_factor = sequence(kwargs.get('heat_loss_factor', 0))
        self.heat_loss_factor_fix = sequence(kwargs.get(
            'heat_loss_factor_fix', 0))

        self._invest_group = False
        self._nonconvex_group = False
        self._demand_group = False

        if len(self.inputs) > 1 or len(self.outputs) > 2:
            if len(self.outputs) == 2:
                self._demand_group = True
            else:
                raise ValueError("Heatpipe must not have more than"
                                 " one input and two outputs!")

        for f in self.inputs.values():
            if f.nonconvex is not None:
                raise ValueError(
                    "Inputflow must not be of type NonConvexFlow!")

        for f in self.outputs.values():
            if f.nonconvex is not None:
                self._nonconvex_group = True

        self._check_flows_invest()

        if (self._nonconvex_group is True) and (self._invest_group is True):
            raise ValueError(
                "Either an investment OR a switchable heatloss can be set"
                " (NonConvexFlow)."
                " Remove the NonConvexFlow or drop "
                "the Investment attribute.")

        if self._invest_group is True:
            self._set_flows()
            o = list(self.outputs.keys())[0]
            if (self.heat_loss_factor_fix[0] > 0) \
                    and (self.outputs[o].investment.nonconvex is False):
                warnings.warn(
                    "Be careful! In case of a convex Investment "
                    "(Investment.nonconvex is False), the "
                    "'heat_loss_factor_fix' is considered, even though the "
                    "investment might be zero! => A simple sink could be "
                    "the results. Hopefully, you know what you are doing.")
        else:
            self._set_nominal_value()

    def _check_flows_invest(self):
        for flow in self.inputs.values():
            if isinstance(flow.investment, Investment):
                raise ValueError(
                    "The investment must be defined at the Outputflow!")

        for flow in self.outputs.values():
            if isinstance(flow.investment, Investment):
                self._invest_group = True

    def _set_flows(self):
        # sets the input flow to investment
        for flow in self.inputs.values():
            flow.investment = Investment()

    # set nominal values of in- and outflow equal in case of invest_group = False
    def _set_nominal_value(self):
        i = list(self.inputs.keys())[0]
        o = list(self.outputs.keys())[0]
        if self.outputs[o].nominal_value is not None:
            self.inputs[i].nominal_value = \
                self.outputs[o].nominal_value
        elif self.inputs[i].nominal_value is not None:
            self.outputs[o].nominal_value = \
                self.inputs[i].nominal_value

    def constraint_group(self):
        if self._invest_group is True:
            return HeatPipelineInvestBlock
        return HeatPipelineBlock


class HeatPipelineBlock(SimpleBlock):  # pylint: disable=too-many-ancestors
    r"""Block representing a pipeline of a district heating system.
    :class:`~oemof.solph.custom.HeatPipeline`

    **The following constraints are created:**

    .. _HeatPipelineBlock-equations:

    .. math::
        &
        (1) \dot{Q}_{out}(t) = \dot{Q}_{in}(t) \cdot
        \frac{\eta_{out}}{\eta_{in}} - \dot{Q}_{loss}(t)\\
        &
        (2) \dot{Q}_{loss}(t) = f_{loss}(t) \cdot l \cdot \dot{Q}_{nominal}
        &

    The symbols used are defined as follows
    (with Variables (V) and Parameters (P)):

    .. csv-table::
        :header: "symbol", "attribute", "type", "explanation"
        :widths: 1, 1, 1, 1

        ":math:`\dot{Q}_{out}(t)`", ":py:obj:`flow[n, o, t]`", "V", "Heat
        output"
        ":math:`\dot{Q}_{in}(t)`", ":py:obj:`flow[i, n, t]`", "V", "Heat input"
        ":math:`\dot{Q}_{loss}(t)`", ":py:obj:`heat_loss[n, t]`", "P", "Heat
        loss of heat pipeline"
        ":math:`\dot{Q}_{nominal}`", ":py:obj:`flows[n, o].nominal_value`", "
        P", "Nominal capacity of heating pipeline"
        ":math:`\eta_{out}`", ":py:obj:`conversion_factors[o][t]`", "P", "
        Conversion factor of output flow (Heat Output)"
        ":math:`\eta_{in}`", ":py:obj:`conversion_factors[i][t]`", "P", "
        Conversion factor of input flow (Heat Input)"
        ":math:`f_{loss}(t)`", ":py:obj:`heat_loss_factor`", "P", "Specific
        heat loss factor for pipeline"
        ":math:`l`", ":py:obj:`length`", "P", "Length of heating pipeline"


    """

    CONSTRAINT_GROUP = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _create(self, group=None):
        """ Creates the linear constraint for the class:`Heatpipe`
        block.

        Parameters
        ----------
        group : list

        """
        if group is None:
            return None

        m = self.parent_block()

        self.HEATPIPES = Set(initialize=list(group))
        self.CONVEX_HEATPIPES = Set(initialize=[
            n for n in group if n.outputs[list(n.outputs.keys())[0]].nonconvex is None])
        self.NONCONVEX_HEATPIPES = Set(initialize=[
            n for n in group if n.outputs[list(n.outputs.keys())[0]].nonconvex is not None])

        # Defining Variables
        self.heat_loss = Var(self.HEATPIPES, m.TIMESTEPS,
                             within=NonNegativeReals)

        def _heat_loss_rule_fix(block, n, t):
            """Rule definition for the heat loss depending on the nominal
            capacity for fix fix heat loss.
            """
            o = list(n.outputs.keys())[0]

            expr = 0
            expr += - block.heat_loss[n, t]
            expr +=\
                n.heat_loss_factor[t] * m.flows[n, o].nominal_value
            expr += n.heat_loss_factor_fix[t]
            return expr == 0

        self.heat_loss_equation_fix = Constraint(
            self.CONVEX_HEATPIPES, m.TIMESTEPS, rule=_heat_loss_rule_fix)

        def _heat_loss_rule_on_off(block, n, t):
            """Rule definition for the heat loss depending on the nominal
            capacity. Here, the losses can be "switched off".
            """
            o = list(n.outputs.keys())[0]

            expr = 0
            expr += - block.heat_loss[n, t]
            expr += \
                (n.heat_loss_factor[t] * m.flows[n, o].nominal_value + n.heat_loss_factor_fix[t]) *\
                m.NonConvexFlow.status[n, o, t]
            return expr == 0

        self.heat_loss_equation_on_off = Constraint(
            self.NONCONVEX_HEATPIPES, m.TIMESTEPS, rule=_heat_loss_rule_on_off)

        def _relation_rule(block, n, t):
            """Link input and output flow and subtract heat loss."""
            i = list(n.inputs.keys())[0]
            o = list(n.outputs.keys())[0]

            expr = 0
            expr += - m.flow[n, o, t]
            expr += m.flow[i, n, t] * n.conversion_factors[
                o][t] / n.conversion_factors[i][t]
            expr += - block.heat_loss[n, t]
            return expr == 0

        self.relation = Constraint(self.HEATPIPES, m.TIMESTEPS,
                                   rule=_relation_rule)


class HeatPipelineInvestBlock(SimpleBlock):  # pylint: disable=too-many-ancestors
    r"""Block representing a pipeline of a district heating system.
    :class:`~oemof.solph.custom.HeatPipeline`

    **The following constraints are created:**

    .. _HeatPipelineInvestBlock-equations:

    .. math::
        &
        (1) \dot{Q}_{out}(t) = \dot{Q}_{in}(t) \cdot
        \frac{\eta_{out}}{\eta_{in}} - \dot{Q}_{loss}(t)\\
        &
        (2) \dot{Q}_{loss}(t) = f_{loss}(t) \cdot l \cdot \dot{Q}_{nominal}
        &

    The symbols used are defined as follows
    (with Variables (V) and Parameters (P)):

    .. csv-table::
        :header: "symbol", "attribute", "type", "explanation"
        :widths: 1, 1, 1, 1

        ":math:`\dot{Q}_{out}(t)`", ":py:obj:`flow[n, o, t]`", "V", "Heat
        output"
        ":math:`\dot{Q}_{in}(t)`", ":py:obj:`flow[i, n, t]`", "V", "Heat input"
        ":math:`\dot{Q}_{loss}(t)`", ":py:obj:`heat_loss[n, t]`", "V", "Heat
        loss of heat pipeline"
        ":math:`\dot{Q}_{nominal}`", ":py:obj:`flows[n, o].nominal_value`", "
        V", "Nominal capacity of heating pipeline"
        ":math:`\eta_{out}`", ":py:obj:`conversion_factors[o][t]`", "P", "
        Conversion factor of output flow (heat output)"
        ":math:`\eta_{in}`", ":py:obj:`conversion_factors[i][t]`", "P", "
        Conversion factor of input flow (heat input)"
        ":math:`f_{loss}(t)`", ":py:obj:`heat_loss_factor`", "P", "Specific
        heat loss factor for pipeline"
        ":math:`l`", ":py:obj:`length`", "P", "Length of heating pipeline"


    """

    CONSTRAINT_GROUP = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _create(self, group=None):
        """ Creates the linear constraint for the class:`HeatPipeline`
        block.

        Parameters
        ----------
        group : list

        """
        if group is None:
            return None

        m = self.parent_block()

        # Defining Sets
        self.INVESTHEATPIPES = Set(initialize=list(group))
        self.CONVEX_INVESTHEATPIPES = Set(initialize=[
            n for n in group if n.outputs[list(n.outputs.keys())[0]].investment.nonconvex is False])
        self.NONCONVEX_INVESTHEATPIPES = Set(initialize=[
            n for n in group if n.outputs[list(n.outputs.keys())[0]].investment.nonconvex is True])

        self.INVESTHEATPIPES_NO_DEMAND = Set(
            initialize=[n for n in group if len(n.outputs.keys()) == 1])
        self.INVESTHEATPIPES_WITH_DEMAND = Set(
            initialize=[n for n in group if len(n.outputs.keys()) == 2])

        # Defining Variables
        self.heat_loss = Var(self.INVESTHEATPIPES, m.TIMESTEPS,
                             within=NonNegativeReals)

        def _heat_loss_rule_convex(block, n, t):
            """Rule definition for constraint to connect the installed capacity
            and the heat loss
            """
            expr = 0
            expr += - block.heat_loss[n, t]
            expr += n.heat_loss_factor[t] * m.InvestmentFlow.invest[n, list(n.outputs.keys())[0]]
            expr += n.heat_loss_factor_fix[t]
            return expr == 0
        self.heat_loss_equation_convex = Constraint(
            self.CONVEX_INVESTHEATPIPES, m.TIMESTEPS, rule=_heat_loss_rule_convex)

        def _heat_loss_rule_nonconvex(block, n, t):
            """Rule definition for constraint to connect the installed capacity
            and the heat loss
            """
            expr = 0
            expr += - block.heat_loss[n, t]
            expr += n.heat_loss_factor[t] * m.InvestmentFlow.invest[n, list(n.outputs.keys())[0]]
            expr += n.heat_loss_factor_fix[t] * \
                m.InvestmentFlow.invest_status[n, list(n.outputs.keys())[0]]
            return expr == 0

        self.heat_loss_equation_nonconvex = Constraint(
            self.NONCONVEX_INVESTHEATPIPES, m.TIMESTEPS, rule=_heat_loss_rule_nonconvex)

        def _relation_rule_no_demand(block, n, t):
            """Link input and output flow and subtract heat loss."""
            i = list(n.inputs.keys())[0]
            o = list(n.outputs.keys())[0]

            expr = 0
            expr += - m.flow[n, o, t]
            expr += m.flow[i, n, t] * n.conversion_factors[
                o][t] / n.conversion_factors[i][t]
            expr += - block.heat_loss[n, t]
            return expr == 0

        self.relation_no_demand = Constraint(
            self.INVESTHEATPIPES_NO_DEMAND, m.TIMESTEPS, rule=_relation_rule_no_demand)

        def _relation_rule_with_demand(block, n, t):
            """Link input and output flow and subtract heat loss."""
            i = list(n.inputs.keys())[0]
            o = list(n.outputs.keys())[0]
            d = list(n.outputs.keys())[1]

            expr = 0
            expr += - m.flow[n, o, t]
            expr += m.flow[i, n, t] * n.conversion_factors[
                o][t] / n.conversion_factors[i][t]
            expr += - block.heat_loss[n, t]
            expr += - m.flow[n, d, t]
            return expr == 0
        self.relation_with_demand = Constraint(
            self.INVESTHEATPIPES_WITH_DEMAND, m.TIMESTEPS, rule=_relation_rule_with_demand)

        def _inflow_outflow_invest_coupling_rule(block, n):  # pylint: disable=unused-argument
            """Rule definition of constraint connecting the inflow
            `InvestmentFlow.invest of pipe with invested outflow `invest`
            by nominal_storage_capacity__inflow_ratio
            """
            i = list(n.inputs.keys())[0]
            o = list(n.outputs.keys())[0]

            expr = (m.InvestmentFlow.invest[i, n] == m.InvestmentFlow.invest[n, o])
            return expr
        self.inflow_outflow_invest_coupling = Constraint(
            self.INVESTHEATPIPES, rule=_inflow_outflow_invest_coupling_rule)



#!/usr/bin/env python

from setuptools import setup
import os


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name='dhnx',
    version='0.0.1',
    description='Simulation and optimization of district heating and cooling networks',
    author="oemof developer group",
    author_email='jann.launer@rl-institut.de, johannes-roeder@uni-bremen.de',
    long_description=read('README.rst'),
    long_description_content_type="text/x-rst",
    packages=['dhnx'],
    package_data={
        'dhnx': [
            "*.csv",
            os.path.join("component_attrs", "*.csv"),
        ],
    },
    install_requires=[
        'pandas >= 0.18.0',
        'matplotlib',
        'networkx',
        'pillow',
        'folium',
        'addict',
        'oemof.solph >= 0.4.0',
        'scipy >= 1.5',
    ],
    extras_require={
        'cartopy': ['cartopy'],
        'geopandas': ['geopandas'],
        'osmnx': ['osmnx'],
    }
)
